#!/usr/bin/python
# -*- coding: utf-8 -*-
"""unittests for acme2certifier"""

# pylint: disable= C0415, W0212
import unittest
import sys
import os
from OpenSSL import crypto
from unittest.mock import patch, Mock, MagicMock, mock_open
import requests
import configparser

sys.path.insert(0, ".")
sys.path.insert(1, "..")


class TestACMEHandler(unittest.TestCase):
    """test class for cgi_handler"""

    def setUp(self):
        """setup unittest"""
        import logging

        logging.basicConfig(level=logging.CRITICAL)
        self.logger = logging.getLogger("test_a2c")
        from acme2certifier.eabhandlers.file_handler import EABhandler

        self.eabhandler = EABhandler(self.logger)
        self.dir_path = os.path.dirname(os.path.realpath(__file__))

    def test_001_default(self):
        """default test which always passes"""
        self.assertEqual("foo", "foo")

    @patch("acme2certifier.eabhandlers.file_handler.EABhandler._config_load")
    def test_002__enter__(self, mock_cfg):
        """test enter  called"""
        mock_cfg.return_value = True
        self.eabhandler.__enter__()
        self.assertTrue(mock_cfg.called)

    @patch("acme2certifier.eabhandlers.file_handler.load_config")
    def test_003_config_load(self, mock_load_cfg):
        """test _config_load - empty dictionary"""
        parser = configparser.ConfigParser()
        mock_load_cfg.return_value = parser
        self.eabhandler._config_load()
        self.assertFalse(self.eabhandler.key_file)

    @patch("acme2certifier.eabhandlers.file_handler.load_config")
    def test_004_config_load(self, mock_load_cfg):
        """test _config_load - no values"""
        parser = configparser.ConfigParser()
        parser["foo"] = {"foo": "bar"}
        mock_load_cfg.return_value = parser
        self.eabhandler._config_load()
        self.assertFalse(self.eabhandler.key_file)

    @patch("acme2certifier.eabhandlers.file_handler.load_config")
    def test_005_config_load(self, mock_load_cfg):
        """test _config_load - bogus values"""
        parser = configparser.ConfigParser()
        parser["EABhandler"] = {"foo": "bar"}
        mock_load_cfg.return_value = parser
        self.eabhandler._config_load()
        self.assertFalse(self.eabhandler.key_file)

    @patch("acme2certifier.eabhandlers.file_handler.load_config")
    def test_006_config_load(self, mock_load_cfg):
        """test _config_load - bogus values"""
        parser = configparser.ConfigParser()
        parser["EABhandler"] = {"key_file": "key_file"}
        mock_load_cfg.return_value = parser
        self.eabhandler._config_load()
        self.assertEqual("key_file", self.eabhandler.key_file)

    def test_007_mac_key_get(self):
        """test mac_key_get without file specified"""
        self.assertFalse(self.eabhandler.mac_key_get(None))

    @patch("builtins.open", mock_open(read_data="foo"), create=True)
    def test_008_mac_key_get(self):
        """test mac_key_get with file but no kid"""
        self.eabhandler.key_file = "file"
        self.assertFalse(self.eabhandler.mac_key_get(None))

    @patch("csv.DictReader")
    @patch("builtins.open", mock_open(read_data="foo"), create=True)
    def test_009_mac_key_get(self, mock_csv):
        """test mac_key_get csv reader return bogus values"""
        self.eabhandler.key_file = "file"
        mock_csv.return_value = ["foo", "bar"]
        self.assertFalse(self.eabhandler.mac_key_get("kid"))

    @patch("csv.DictReader")
    @patch("builtins.open", mock_open(read_data="foo"), create=True)
    def test_010_mac_key_get(self, mock_csv):
        """test mac_key_get csv reader return match"""
        self.eabhandler.key_file = "file"
        mock_csv.return_value = [{"eab_kid": "kid", "eab_mac": "mac"}]
        self.assertEqual("mac", self.eabhandler.mac_key_get("kid"))

    @patch("csv.DictReader")
    @patch("builtins.open", mock_open(read_data="foo"), create=True)
    def test_011_mac_key_get(self, mock_csv):
        """test mac_key_get csv reader no match"""
        self.eabhandler.key_file = "file"
        mock_csv.return_value = [{"eab_kid": "kid1", "eab_mac": "mac"}]
        self.assertFalse(self.eabhandler.mac_key_get("kid"))

    @patch("csv.DictReader")
    @patch("builtins.open", mock_open(read_data="foo"), create=True)
    def test_012_mac_key_get(self, mock_csv):
        """test mac_key_get check break after first match"""
        self.eabhandler.key_file = "file"
        mock_csv.return_value = [
            {"eab_kid": "kid1", "eab_mac": "mac1"},
            {"eab_kid": "kid2", "eab_mac": "mac2"},
        ]
        self.assertEqual("mac1", self.eabhandler.mac_key_get("kid1"))

    @patch("csv.DictReader")
    @patch("builtins.open", mock_open(read_data="foo"), create=True)
    def test_013_mac_key_get(self, mock_csv):
        """test mac_key_get match in the 2nd record"""
        self.eabhandler.key_file = "file"
        mock_csv.return_value = [
            {"eab_kid": "kid1", "eab_mac": "mac1"},
            {"eab_kid": "kid2", "eab_mac": "mac2"},
        ]
        self.assertEqual("mac2", self.eabhandler.mac_key_get("kid2"))

    @patch("csv.DictReader")
    @patch("builtins.open", mock_open(read_data="foo"), create=True)
    def test_014_mac_key_get(self, mock_csv):
        """test mac_key_get csv reader no eab_kid"""
        self.eabhandler.key_file = "file"
        mock_csv.return_value = [{"eab__kid": "kid", "eab_mac": "mac"}]
        self.assertFalse(self.eabhandler.mac_key_get("kid"))

    @patch("csv.DictReader")
    @patch("builtins.open", mock_open(read_data="foo"), create=True)
    def test_015_mac_key_get(self, mock_csv):
        """test mac_key_get csv reader no mac but match"""
        self.eabhandler.key_file = "file"
        mock_csv.return_value = [{"eab_kid": "kid", "_eab_mac": "mac"}]
        self.assertFalse(self.eabhandler.mac_key_get("kid"))

    @patch("csv.DictReader")
    @patch("builtins.open", mock_open(read_data="foo"), create=True)
    def test_016_mac_key_get(self, mock_csv):
        """test mac_key_get csv reader no mac but match"""
        self.eabhandler.key_file = "file"
        mock_csv.side_effect = Exception("ex_mock_csv")
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertFalse(self.eabhandler.mac_key_get("kid"))
        self.assertIn(
            "ERROR:test_a2c:Failed to load EAB key file: ex_mock_csv", lcm.output
        )


if __name__ == "__main__":

    unittest.main()



"""
Enhanced error handling system for challenge processing.

This module provides a comprehensive error handling framework with custom exceptions,
error categorization, and standardized error responses for challenge operations.
"""

from typing import Dict, Optional, Any, List
from dataclasses import dataclass
from enum import Enum
import traceback
import logging


class ErrorCategory(Enum):
    """Categories of errors that can occur during challenge processing."""

    VALIDATION_ERROR = "validation_error"
    NETWORK_ERROR = "network_error"
    DATABASE_ERROR = "database_error"
    CONFIGURATION_ERROR = "configuration_error"
    AUTHENTICATION_ERROR = "authentication_error"
    MALFORMED_REQUEST = "malformed_request"
    TIMEOUT_ERROR = "timeout_error"
    UNKNOWN_ERROR = "unknown_error"


class ErrorSeverity(Enum):
    """Severity levels for errors."""

    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"
    CRITICAL = "critical"


@dataclass
class ErrorDetail:
    """Detailed error information."""

    category: ErrorCategory
    severity: ErrorSeverity
    message: str
    details: Optional[Dict[str, Any]] = None
    suggestion: Optional[str] = None
    error_code: Optional[str] = None


class ChallengeError(Exception):
    """Base exception for all challenge-related errors."""

    def __init__(
        self,
        message: str,
        category: ErrorCategory = ErrorCategory.UNKNOWN_ERROR,
        severity: ErrorSeverity = ErrorSeverity.MEDIUM,
        details: Optional[Dict[str, Any]] = None,
        suggestion: Optional[str] = None,
        error_code: Optional[str] = None,
    ):
        super().__init__(message)
        self.error_detail = ErrorDetail(
            category=category,
            severity=severity,
            message=message,
            details=details or {},
            suggestion=suggestion,
            error_code=error_code,
        )


class ValidationError(ChallengeError):
    """Raised when challenge validation fails."""

    def __init__(self, message: str, **kwargs):
        super().__init__(message, category=ErrorCategory.VALIDATION_ERROR, **kwargs)


class NetworkError(ChallengeError):
    """Raised when network operations fail."""

    def __init__(self, message: str, **kwargs):
        super().__init__(message, category=ErrorCategory.NETWORK_ERROR, **kwargs)


class DatabaseError(ChallengeError):
    """Raised when database operations fail."""

    def __init__(self, message: str, **kwargs):
        super().__init__(
            message,
            category=ErrorCategory.DATABASE_ERROR,
            severity=ErrorSeverity.HIGH,
            **kwargs,
        )


class ConfigurationError(ChallengeError):
    """Raised when configuration is invalid."""

    def __init__(self, message: str, **kwargs):
        super().__init__(
            message,
            category=ErrorCategory.CONFIGURATION_ERROR,
            severity=ErrorSeverity.HIGH,
            **kwargs,
        )


class AuthenticationError(ChallengeError):
    """Raised when authentication fails."""

    def __init__(self, message: str, **kwargs):
        super().__init__(
            message,
            category=ErrorCategory.AUTHENTICATION_ERROR,
            severity=ErrorSeverity.HIGH,
            **kwargs,
        )


class MalformedRequestError(ChallengeError):
    """Raised when request is malformed."""

    def __init__(self, message: str, **kwargs):
        super().__init__(message, category=ErrorCategory.MALFORMED_REQUEST, **kwargs)


class TimeoutError(ChallengeError):
    """Raised when operations timeout."""

    def __init__(self, message: str, **kwargs):
        super().__init__(message, category=ErrorCategory.TIMEOUT_ERROR, **kwargs)


class UnsupportedChallengeTypeError(ValidationError):
    """Raised when an unsupported challenge type is encountered."""

    def __init__(self, challenge_type: str, supported_types: List[str]):
        message = f"Unsupported challenge type: {challenge_type}"
        super().__init__(
            message,
            details={
                "challenge_type": challenge_type,
                "supported_types": supported_types,
            },
            suggestion=f"Use one of the supported types: {', '.join(supported_types)}",
            error_code="UNSUPPORTED_CHALLENGE_TYPE",
        )


class DNSResolutionError(NetworkError):
    """Raised when DNS resolution fails."""

    def __init__(self, domain: str, dns_servers: Optional[List[str]] = None):
        message = f"DNS resolution failed for domain: {domain}"
        super().__init__(
            message,
            details={"domain": domain, "dns_servers": dns_servers},
            suggestion="Check domain validity and DNS server configuration",
            error_code="DNS_RESOLUTION_FAILED",
        )


class HTTPChallengeError(ValidationError):
    """Raised when HTTP challenge validation fails."""

    def __init__(self, url: str, expected: str, received: str):
        message = f"HTTP challenge validation failed for {url}"
        super().__init__(
            message,
            details={
                "url": url,
                "expected_response": expected,
                "received_response": received,
            },
            suggestion="Ensure the challenge file is accessible and contains the correct token",
            error_code="HTTP_CHALLENGE_FAILED",
        )


class DNSChallengeError(ValidationError):
    """Raised when DNS challenge validation fails."""

    def __init__(self, dns_record: str, expected_hash: str, found_records: List[str]):
        message = f"DNS challenge validation failed for {dns_record}"
        super().__init__(
            message,
            details={
                "dns_record": dns_record,
                "expected_hash": expected_hash,
                "found_records": found_records,
            },
            suggestion="Ensure the DNS TXT record is properly configured",
            error_code="DNS_CHALLENGE_FAILED",
        )


class TLSALPNChallengeError(ValidationError):
    """Raised when TLS-ALPN challenge validation fails."""

    def __init__(self, domain: str, expected_extension: str):
        message = f"TLS-ALPN challenge validation failed for {domain}"
        super().__init__(
            message,
            details={"domain": domain, "expected_extension": expected_extension},
            suggestion="Ensure the TLS certificate contains the required extension",
            error_code="TLS_ALPN_CHALLENGE_FAILED",
        )


class ErrorHandler:
    """Centralized error handling and logging."""

    def __init__(self, logger: logging.Logger):
        self.logger = logger
        self.error_counts: Dict[ErrorCategory, int] = {}

    def handle_error(
        self, error: Exception, context: Optional[Dict[str, Any]] = None
    ) -> ErrorDetail:
        """Handle and log an error, returning structured error information."""
        self.logger.debug("ErrorHandler.handle_error(): %s", str(error))
        if isinstance(error, ChallengeError):
            error_detail = error.error_detail
        else:
            # Convert generic exceptions to ChallengeError
            error_detail = ErrorDetail(
                category=ErrorCategory.UNKNOWN_ERROR,
                severity=ErrorSeverity.MEDIUM,
                message=str(error),
                details={"exception_type": type(error).__name__},
            )

        # Add context information
        if context:
            error_detail.details.update(context)

        # Log the error
        self._log_error(error_detail, error)

        # Update error counts
        # self._update_error_counts(error_detail.category)

        return error_detail

    def _log_error(self, error_detail: ErrorDetail, original_error: Exception):
        """Log error with appropriate level based on severity."""
        self.logger.debug("ErrorHandler._log_error(): %s", str(original_error))
        log_message = f"[{error_detail.category.value}] {error_detail.message}"

        if error_detail.details:
            log_message += f" | Details: {error_detail.details}"

        if error_detail.severity == ErrorSeverity.CRITICAL:
            self.logger.critical(log_message, exc_info=True)
        elif error_detail.severity == ErrorSeverity.HIGH:
            self.logger.error(log_message)
        elif error_detail.severity == ErrorSeverity.MEDIUM:
            self.logger.warning(log_message)
        else:
            self.logger.info(log_message)

        # Log stack trace for debugging in debug mode
        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug(
                "Stack trace for error: %s",
                "".join(
                    traceback.format_exception(
                        type(original_error),
                        original_error,
                        original_error.__traceback__,
                    )
                ),
            )

    def create_acme_error_response(
        self, error_detail: ErrorDetail, status_code: int = 400
    ) -> Dict[str, Any]:
        """Create an ACME-compliant error response."""

        # Map internal categories to ACME error types
        acme_error_type_map = {
            ErrorCategory.VALIDATION_ERROR: "incorrectResponse",
            ErrorCategory.NETWORK_ERROR: "connection",
            ErrorCategory.MALFORMED_REQUEST: "malformed",
            ErrorCategory.AUTHENTICATION_ERROR: "unauthorized",
            ErrorCategory.TIMEOUT_ERROR: "connection",
            ErrorCategory.CONFIGURATION_ERROR: "serverInternal",
            ErrorCategory.DATABASE_ERROR: "serverInternal",
            ErrorCategory.UNKNOWN_ERROR: "serverInternal",
        }

        acme_type = acme_error_type_map.get(error_detail.category, "serverInternal")

        response = {
            "code": status_code,
            "type": f"urn:ietf:params:acme:error:{acme_type}",
            "detail": error_detail.message,
        }

        # Add additional context if available
        if error_detail.suggestion:
            response["detail"] += f" Suggestion: {error_detail.suggestion}"

        return response


class ErrorRecovery:
    """Provides error recovery strategies."""

    def __init__(self, logger: logging.Logger):
        self.logger = logger

    def should_retry(self, error_detail: ErrorDetail, attempt_count: int) -> bool:
        """Determine if an operation should be retried based on error type."""

        # Don't retry beyond maximum attempts
        if attempt_count >= 3:
            return False

        # Retry network errors and timeouts
        if error_detail.category in [
            ErrorCategory.NETWORK_ERROR,
            ErrorCategory.TIMEOUT_ERROR,
        ]:
            return True

        # Don't retry validation errors, malformed requests, or authentication errors
        if error_detail.category in [
            ErrorCategory.VALIDATION_ERROR,
            ErrorCategory.MALFORMED_REQUEST,
            ErrorCategory.AUTHENTICATION_ERROR,
        ]:
            return False

        # Retry database errors (might be temporary)
        if error_detail.category == ErrorCategory.DATABASE_ERROR:
            return True

        return False

    def get_retry_delay(self, attempt_count: int) -> float:
        """Get delay before retry with exponential backoff."""
        return min(2**attempt_count, 30)  # Max 30 seconds



# -*- coding: utf-8 -*-
"""acme app main view"""

from __future__ import unicode_literals, print_function
from django.http import HttpResponse, HttpResponseNotFound
from django.utils.html import escape
from acme2certifier.django_app.a2c_response import JsonResponse
from acme2certifier.acme_srv.authorization import Authorization
from acme2certifier.acme_srv.account import Account
from acme2certifier.acme_srv.certificate import Certificate
from acme2certifier.acme_srv.challenge import Challenge
from acme2certifier.acme_srv.directory import Directory
from acme2certifier.acme_srv.helper import (
    get_url,
    load_config,
    logger_setup,
    logger_info,
    config_check,
)
from acme2certifier.acme_srv.db_handler import log_active_db_handler
from acme2certifier.acme_srv.housekeeping import Housekeeping
from acme2certifier.acme_srv.nonce import Nonce
from acme2certifier.acme_srv.order import Order
from acme2certifier.acme_srv.renewalinfo import Renewalinfo
from acme2certifier.acme_srv.trigger import Trigger
from acme2certifier.acme_srv.version import __dbversion__, __version__
from acme2certifier.acme_srv.acmechallenge import Acmechallenge

# load config to set debug mode
CONFIG = load_config()
DEBUG = CONFIG.getboolean("DEFAULT", "debug", fallback=False)

# initialize logger
LOGGER = logger_setup(DEBUG)
LOGGER.info("starting acme2certifier version %s", __version__)
log_active_db_handler(LOGGER)

METHOD_NOT_ALLOWED = "Method Not Allowed"
ERR_DATA_POST = {
    "status": 405,
    "message": METHOD_NOT_ALLOWED,
    "detail": "Wrong request type. Expected POST.",
}
ERR_RESPONSE_POST = JsonResponse(status=405, data=ERR_DATA_POST)
ERR_RESPONSE_HEAD_GET = JsonResponse(
    status=400,
    data={
        "status": 405,
        "message": METHOD_NOT_ALLOWED,
        "detail": "Wrong request type. Expected HEAD or GET.",
    },
)

# check configuration for parameters masked in ""
config_check(LOGGER, CONFIG)

with Housekeeping(DEBUG, LOGGER) as housekeeping:
    housekeeping.dbversion_check(__dbversion__)
    housekeeping.nonce_cleanup()


def handle_exception(exc_type, exc_value, exc_traceback):
    """exception handler"""
    print("My Error Information")
    print("Type:", exc_type)
    print("Value:", exc_value)
    print("Traceback:", exc_traceback)


def pretty_request(request):
    """print request details for debugging"""
    headers = ""
    for header, value in request.META.items():
        if not header.startswith("HTTP"):
            continue
        header = "-".join([h.capitalize() for h in header[5:].lower().split("_")])
        headers += f"{header}: {value}\n"

    return (
        f"{request.method} HTTP/1.1\n"
        f'Content-Length: {request.META["CONTENT_LENGTH"]}\n'
        f'Content-Type: {request.META["CONTENT_TYPE"]}\n'
        f"{headers}\n\n"
        f"{request.body}"
    )


def directory(request):
    """get directory"""
    with Directory(DEBUG, get_url(request.META), LOGGER) as cfg_dir:
        response = cfg_dir.directory_get()
        if "error" in response:
            return JsonResponse(
                status=403,
                data={
                    "status": 403,
                    "message": "Forbidden",
                    "detail": response["error"],
                },
            )
        else:
            return JsonResponse(cfg_dir.directory_get())


def newaccount(request):
    """new account"""
    if request.method == "POST":
        with Account(DEBUG, get_url(request.META), LOGGER) as account:
            response_dic = account.new(request.body)
            # create the response
            response = JsonResponse(
                status=response_dic["code"], data=response_dic["data"]
            )

            # generate additional header elements
            for element in response_dic["header"]:
                response[element] = response_dic["header"][element]

            # logging
            logger_info(
                LOGGER,
                request.META["REMOTE_ADDR"],
                request.META["PATH_INFO"],
                response_dic,
            )
            # send response
            return response
    else:
        return ERR_RESPONSE_POST


def newnonce(request):
    """new nonce"""
    if request.method in ["HEAD", "GET"]:
        with Nonce(DEBUG, LOGGER) as nonce:
            if request.method == "HEAD":
                response = HttpResponse("")
            else:
                response = HttpResponse(status=204)

            # do housekeeping and expire old nonces
            nonce.expire_nonces()

            # generate nonce
            response["Replay-Nonce"] = nonce.generate_and_add()

            # logging
            logger_info(
                LOGGER,
                request.META["REMOTE_ADDR"],
                request.META["PATH_INFO"],
                {"header": {"Replay-Nonce": response["Replay-Nonce"]}},
            )
            # send response
            return response
    else:
        return ERR_RESPONSE_HEAD_GET


def servername_get(request):
    """get server name"""
    with Directory(DEBUG, get_url(request.META), LOGGER) as cfg_dir:
        return JsonResponse({"server_name": escape(cfg_dir.servername_get())})


def acct(request):
    """xxxx command"""
    with Account(DEBUG, get_url(request.META), LOGGER) as account:
        response_dic = account.parse(request.body)
        # create the response
        response = JsonResponse(status=response_dic["code"], data=response_dic["data"])

        # generate additional header elements
        for element in response_dic["header"]:
            response[element] = response_dic["header"][element]

        # logging
        logger_info(
            LOGGER, request.META["REMOTE_ADDR"], request.META["PATH_INFO"], response_dic
        )
        # send response
        return response


def neworders(request):
    """new account"""
    if request.method == "POST":
        with Order(DEBUG, get_url(request.META), LOGGER) as norder:
            response_dic = norder.new(request.body)
            # create the response
            response = JsonResponse(
                status=response_dic["code"], data=response_dic["data"]
            )

            # generate additional header elements
            for element in response_dic["header"]:
                response[element] = response_dic["header"][element]

            if "Replay-Nonce" not in response:
                response["Replay-Nonce"] = ""

            # logging
            logger_info(
                LOGGER,
                request.META["REMOTE_ADDR"],
                request.META["PATH_INFO"],
                {"header": {"Replay-Nonce": response["Replay-Nonce"]}},
            )
            # send response
            return response
    else:
        return ERR_RESPONSE_POST


def authz(request):
    """new-authz command"""
    if request.method in ("POST", "GET"):
        with Authorization(DEBUG, get_url(request.META), LOGGER) as authorization:
            if request.method == "POST":
                response_dic = authorization.new_post(request.body)
            else:
                response_dic = authorization.new_get(request.build_absolute_uri())
            # create the response
            response = JsonResponse(
                status=response_dic["code"], data=response_dic["data"]
            )

            # generate additional header elements
            for element in response_dic["header"]:
                response[element] = response_dic["header"][element]

            # logging
            logger_info(
                LOGGER,
                request.META["REMOTE_ADDR"],
                request.META["PATH_INFO"],
                response_dic,
            )
            # send response
            return response
    else:
        return ERR_RESPONSE_POST


def chall(request):
    """challenge command"""
    with Challenge(
        debug=DEBUG,
        srv_name=get_url(request.META),
        source=request.META["REMOTE_ADDR"],
        logger=LOGGER,
    ) as challenge:
        # pylint: disable=R1705
        if request.method == "POST":
            response_dic = challenge.parse(request.body)
            # create the response
            response = JsonResponse(
                status=response_dic["code"], data=response_dic["data"]
            )
            # generate additional header elements
            for element in response_dic["header"]:
                response[element] = response_dic["header"][element]

            # logging
            logger_info(
                LOGGER,
                request.META["REMOTE_ADDR"],
                request.META["PATH_INFO"],
                response_dic,
            )
            # send response
            return response
        elif request.method == "GET":
            response_dic = challenge.get(request.build_absolute_uri())
            # create the response
            response = JsonResponse(
                status=response_dic["code"], data=response_dic["data"]
            )

            # send response
            return response
        else:
            return ERR_RESPONSE_POST


def order(request):
    """order request"""
    if request.method == "POST":
        with Order(DEBUG, get_url(request.META), LOGGER) as eorder:
            response_dic = eorder.parse(request.body, request.META)
            # create the response
            response = JsonResponse(
                status=response_dic["code"], data=response_dic["data"]
            )
            # generate additional header elements
            for element in response_dic["header"]:
                response[element] = response_dic["header"][element]

            # logging
            logger_info(
                LOGGER,
                request.META["REMOTE_ADDR"],
                request.META["PATH_INFO"],
                response_dic,
            )
            # send response
            return response
    else:
        return ERR_RESPONSE_POST


def cert(request):
    """cert request"""
    if request.method in ("POST", "GET"):
        with Certificate(DEBUG, get_url(request.META), LOGGER) as certificate:
            if request.method == "POST":
                response_dic = certificate.new_post(request.body)
            else:
                response_dic = certificate.new_get(request.build_absolute_uri())

            # create the response
            if response_dic["code"] == 200:
                response = HttpResponse(response_dic["data"])
                # generate additional header elements
                for element in response_dic["header"]:
                    response[element] = response_dic["header"][element]
            else:
                response = HttpResponse(status=response_dic["code"])

            # logging
            logger_info(
                LOGGER,
                request.META["REMOTE_ADDR"],
                request.META["PATH_INFO"],
                response_dic,
            )
            # send response
            return response

    else:
        return ERR_RESPONSE_POST


def revokecert(request):
    """cert revocation"""
    if request.method == "POST":
        with Certificate(DEBUG, get_url(request.META), LOGGER) as certificate:
            response_dic = certificate.revoke(request.body)
            # create the response
            if "data" in response_dic:
                response = JsonResponse(
                    status=response_dic["code"], data=response_dic["data"]
                )
            else:
                response = HttpResponse(status=response_dic["code"])

            # generate additional header elements
            for element in response_dic["header"]:
                response[element] = response_dic["header"][element]

            # logging
            logger_info(
                LOGGER,
                request.META["REMOTE_ADDR"],
                request.META["PATH_INFO"],
                response_dic,
            )
            # send response
            return response
    else:
        return ERR_RESPONSE_POST


def trigger(request):
    """ca trigger"""
    if request.method == "POST":
        with Trigger(DEBUG, get_url(request.META), LOGGER) as trigger_:
            response_dic = trigger_.parse(request.body)
            # create the response
            if "data" in response_dic:
                response = JsonResponse(
                    status=response_dic["code"], data=response_dic["data"]
                )
            else:
                response = HttpResponse(status=response_dic["code"])

            # generate additional header elements
            for element in response_dic["header"]:
                response[element] = response_dic["header"][element]

            # logging
            logger_info(
                LOGGER,
                request.META["REMOTE_ADDR"],
                request.META["PATH_INFO"],
                response_dic,
            )
            # send response
            return response
    else:
        return ERR_RESPONSE_POST


def renewalinfo(request):
    """renewal info"""
    if request.method in ("POST", "GET"):
        with Renewalinfo(DEBUG, get_url(request.META), LOGGER) as renewalinfo_:
            if request.method == "POST":
                response_dic = renewalinfo_.update(request.body)
            else:
                response_dic = renewalinfo_.get(request.build_absolute_uri())

            # create the response
            if response_dic["code"] == 200 and request.method == "GET":
                response = JsonResponse(response_dic["data"])
                # generate additional header elements
                for element in response_dic["header"]:
                    response[element] = response_dic["header"][element]
            else:
                response = HttpResponse(status=response_dic["code"])

            # logging
            logger_info(
                LOGGER,
                request.META["REMOTE_ADDR"],
                request.META["PATH_INFO"],
                response_dic,
            )

            # send response
            return response

    else:
        return ERR_RESPONSE_POST


def housekeeping(request):
    """ca trigger"""
    if request.method == "POST":
        with Housekeeping(DEBUG, LOGGER) as housekeeping_:
            response_dic = housekeeping_.parse(request.body)
            # create the response
            if "data" in response_dic:
                response = JsonResponse(
                    status=response_dic["code"], data=response_dic["data"], safe=False
                )
            else:
                response = HttpResponse(status=response_dic["code"])

            # generate additional header elements
            for element in response_dic["header"]:
                response[element] = response_dic["header"][element]

            # logging
            logger_info(
                LOGGER, request.META["REMOTE_ADDR"], request.META["PATH_INFO"], "****"
            )
            # send response
            return response
    else:
        return JsonResponse(status=405, data=ERR_DATA_POST, safe=False)


def acmechallenge_serve(request):
    """serving acme challenges"""
    with Acmechallenge(DEBUG, get_url(request.META), LOGGER) as acmechallenge:
        key_authorization = acmechallenge.lookup(request.META["PATH_INFO"])
        # pylint:  disable=R1705
        if key_authorization:
            return HttpResponse(key_authorization)
        else:
            return HttpResponseNotFound("NOT FOUND")



"""
TKAuth Challenge Validator.

Implements validation logic for tkauth-01 challenges.
"""

from .base import ChallengeValidator, ChallengeContext, ValidationResult


class TkauthChallengeValidator(ChallengeValidator):
    """Validator for tkauth-01 challenges."""

    def get_challenge_type(self) -> str:
        return "tkauth-01"

    def perform_validation(self, context: ChallengeContext) -> ValidationResult:
        """Perform tkauth-01 challenge validation."""
        # For now, this always returns success as in the original implementation
        # This would be expanded with actual validation logic when requirements are defined

        self.logger.debug(
            "TKAuth validation for challenge %s with authorization value %s",
            context.challenge_name,
            context.authorization_value,
        )

        return ValidationResult(
            success=True,
            invalid=False,
            details={
                "validation_type": "tkauth-01",
                "authorization_value": context.authorization_value,
            },
        )



# -*- coding: utf-8 -*-
"""Order class"""

from __future__ import print_function
import json
import copy
from typing import Any, List, Tuple, Dict, Optional
from dataclasses import dataclass, field
from acme2certifier.acme_srv.helper import (
    b64_url_recode,
    ca_handler_load,
    config_allowed_domainlist_load,
    config_allowed_iplist_load,
    config_profile_load,
    error_dic_get,
    generate_random_string,
    load_config,
    parse_url,
    uts_to_date_utc,
    uts_now,
    validate_identifier,
    is_domain_whitelisted,
    config_eab_profile_load,
    config_dryrun_load,
    is_ip_whitelisted,
)
from acme2certifier.acme_srv.certificate import Certificate
from acme2certifier.acme_srv.db_handler import DBstore
from acme2certifier.acme_srv.helpers.global_variables import (
    DRYRUN_ENROLLMENT_SKIPPED_DETAIL,
    ENROLLMENT_FAILED_DETAIL,
)
from acme2certifier.acme_srv.message import Message


class OrderDatabaseError(Exception):
    """Exception raised for database-related errors in Order operations."""

    # pylint: disable=unnecessary-pass
    pass


class OrderValidationError(Exception):
    """Exception raised for validation errors in Order operations."""

    # pylint: disable=unnecessary-pass
    pass


class OrderRepository:
    """Repository for all Order-related database operations."""

    def __init__(self, dbstore, logger):
        self.dbstore = dbstore
        self.logger = logger

    def add_order(self, data_dic):
        """Add a new order to the database."""
        try:
            return self.dbstore.order_add(data_dic)
        except Exception as err:
            self.logger.critical("Database error: failed to add order: %s", err)
            raise OrderDatabaseError(f"Failed to add order: {err}") from err

    def add_authorization(self, auth):
        """Add a new authorization to the database."""
        try:
            return self.dbstore.authorization_add(auth)
        except Exception as err:
            self.logger.critical("Database error: failed to add authorization: %s", err)
            raise OrderDatabaseError(f"Failed to add authorization: {err}") from err

    def update_authorization(self, auth):
        """Update an existing authorization in the database."""
        try:
            return self.dbstore.authorization_update(auth)
        except Exception as err:
            self.logger.critical(
                "Database error: failed to update authorization: %s", err
            )
            raise OrderDatabaseError(f"Failed to update authorization: {err}") from err

    def order_lookup(self, key, value, vlist=None):
        """Look up an order in the database."""
        try:
            if vlist:
                return self.dbstore.order_lookup(key, value, vlist)
            return self.dbstore.order_lookup(key, value)
        except Exception as err:
            self.logger.critical("Database error: failed to look up order: %s", err)
            raise OrderDatabaseError(f"Failed to look up order: {err}") from err

    def order_update(self, data_dic):
        """Update an existing order in the database."""
        try:
            return self.dbstore.order_update(data_dic)
        except Exception as err:
            self.logger.critical("Database error: failed to update order: %s", err)
            raise OrderDatabaseError(f"Failed to update order: {err}") from err

    def authorization_lookup(self, key, value, fields):
        """Look up an authorization in the database."""
        try:
            return self.dbstore.authorization_lookup(key, value, fields)
        except Exception as err:
            self.logger.critical(
                "Database error: failed to look up authorization: %s", err
            )
            raise OrderDatabaseError(f"Failed to look up authorization: {err}") from err

    def account_lookup(self, key, value):
        """Look up an account in the database."""
        try:
            return self.dbstore.account_lookup(key, value)
        except Exception as err:
            self.logger.critical("Database error: failed to look up account: %s", err)
            raise OrderDatabaseError(f"Failed to look up account: {err}") from err

    def certificate_lookup(self, key, value):
        """Look up a certificate in the database."""
        try:
            return self.dbstore.certificate_lookup(key, value)
        except Exception as err:
            self.logger.critical(
                "Database error: failed to look up certificate: %s", err
            )
            raise OrderDatabaseError(f"Failed to look up certificate: {err}") from err

    def hkparameter_get(self, param):
        """Get a hkparameter from the database."""
        try:
            return self.dbstore.hkparameter_get(param)
        except Exception as err:
            self.logger.critical("Database error: failed to get hkparameter: %s", err)
            raise OrderDatabaseError(f"Failed to get hkparameter: {err}") from err

    def orders_invalid_search(self, order_field, timestamp, vlist, operant):
        """Search for invalid orders in the database."""
        try:
            return self.dbstore.orders_invalid_search(
                order_field, timestamp, vlist=vlist, operant=operant
            )
        except Exception as err:
            self.logger.critical(
                "Database error: failed to search for invalid orders: %s", err
            )
            raise OrderDatabaseError(
                f"Failed to search for invalid orders: {err}"
            ) from err


@dataclass
class OrderConfiguration:
    """Configuration parameters for Order handling"""

    validity: int = 86400
    authz_validity: int = 86400
    expiry_check_disable: bool = False
    retry_after: int = 600
    tnauthlist_support: bool = False
    email_identifier_support: bool = False
    email_identifier_rewrite: bool = False
    sectigo_sim: bool = False
    identifier_limit: int = 20
    header_info_list: List[Any] = field(default_factory=list)
    profiles: Dict[str, Any] = field(default_factory=dict)
    profiles_sync: bool = False
    profiles_check_disable: bool = True
    idempotent_finalize: bool = False
    allowed_domainlist: List[str] = field(default_factory=list)
    allowed_iplist: List[str] = field(default_factory=list)
    eab_profiling: bool = False
    eab_handler: Optional[Any] = None
    profile_mapping_field: Optional[str] = None
    dryrun_profilename: Optional[str] = None
    wildcard_certificate_disable: bool = False
    ca_error_details_forward: bool = False


class Order(object):
    """class for order handling"""

    def __init__(
        self, debug: bool = None, server_name: str = None, logger: object = None
    ) -> None:
        """Initialize the Order handler"""
        self.debug = debug
        self.server_name = server_name
        self.config = OrderConfiguration()
        self.logger = logger
        self.dbstore = DBstore(self.debug, self.logger)
        self.path_dic = {
            "authz_path": "/acme/authz/",
            "order_path": "/acme/order/",
            "cert_path": "/acme/cert/",
        }
        self.repository = OrderRepository(self.dbstore, self.logger)
        self.message = Message(self.debug, self.server_name, self.logger)
        self.error_msg_dic = error_dic_get(self.logger)

    def __enter__(self) -> "Order":
        """Enter the context manager, loading configuration."""
        self._load_configuration()
        return self

    def __exit__(self, *args) -> None:
        """
        Exit the context manager. (No-op, placeholder for cleanup.)
        """

    def _add_authorizations_to_db(
        self, oid: str, payload: Dict[str, str], auth_dic: Dict[str, str]
    ) -> str:
        """Add authorizations to the database for the given order id. Returns error message or None."""
        self.logger.debug("Order._add_authorizations_to_db(%s)", oid)

        if oid:
            error = None
            for auth in payload["identifiers"]:
                auth_name = generate_random_string(self.logger, 12)
                auth_dic[auth_name] = auth.copy()
                auth["name"] = auth_name
                auth["order"] = oid
                auth["status"] = "pending"
                auth["expires"] = uts_now() + self.config.authz_validity
                try:
                    self.repository.add_authorization(auth)
                    if self.config.sectigo_sim:
                        auth["status"] = "valid"
                        self.repository.update_authorization(auth)
                except Exception as err_:
                    self.logger.critical(
                        "Database error: failed to add authorization: %s", err_
                    )
                    error = self.error_msg_dic["serverinternal"]
                    break
        else:
            error = self.error_msg_dic["malformed"]

        self.logger.debug("Order._add_authorizations_to_db() ended with %s", error)
        return error

    def is_profile_valid(self, profile: str) -> str:
        """Check if the given profile is valid."""
        self.logger.debug("Order.is_profile_valid(%s)", profile)
        error = self.error_msg_dic["invalidprofile"]
        if self.config.profiles_check_disable:
            self.logger.debug("Order.is_profile_valid(): profile check disabled")
            error = None
        elif (
            self.config.dryrun_profilename and profile == self.config.dryrun_profilename
        ):
            self.logger.info(
                "Order.is_profile_valid(): dryrun profile '%s' submitted",
                profile,
            )
            error = None
        else:
            if profile in self.config.profiles:
                error = None
            else:
                self.logger.warning(
                    "Profile '%s' is not valid. Ignoring submitted profile.", profile
                )
                # cover cornercase, where we have only one profile configured, and the submitted profile is not valid
                # in this case we overwrite the submitted profile silently
                if len(self.config.profiles.keys()) == 1:
                    self.logger.info(
                        "Only one profile configured. Overwriting submitted profile '%s' with '%s'.",
                        profile,
                        next(iter(self.config.profiles.keys())),
                    )
                    error = None

        self.logger.debug("Order.is_profile_valid() ended with %s", error)
        return error

    def _add_order_and_authorizations(
        self,
        data_dic: Dict[str, str],
        auth_dic: Dict[str, str],
        payload: Dict[str, str],
        error: Optional[str] = None,
    ) -> Tuple[str, Dict[str, str]]:
        """Add order and its authorizations to the database. Returns error message or None."""
        self.logger.debug("Order._add_order_and_authorizations()")

        try:
            oid = self.repository.add_order(data_dic)
        except Exception as err_:
            self.logger.critical("Database error: failed to add order: %s", err_)
            oid = None

        if not error:
            error = self._add_authorizations_to_db(oid, payload, auth_dic)

        self.logger.debug("Order._add_order_and_authorizations() ended with %s", error)
        return error

    def add_profile_to_order(
        self, data_dic: Dict[str, str], payload: Dict[str, str]
    ) -> Tuple[str, Dict[str, str]]:
        """Add a profile to the order if valid."""
        self.logger.debug("Order.add_profile_to_order(%s)", data_dic)
        error = self.is_profile_valid(payload["profile"])
        if not error:
            if self.config.profiles:
                profile = payload["profile"]
                if (
                    not self.config.profiles_check_disable
                    and profile not in self.config.profiles
                    and not (
                        self.config.dryrun_profilename
                        and self.config.dryrun_profilename == profile
                    )
                    and len(self.config.profiles.keys()) == 1
                ):
                    profile = next(iter(self.config.profiles.keys()))
                    self.logger.info(
                        "Order.add_profile_to_order(): overwriting submitted profile '%s' with '%s'.",
                        payload["profile"],
                        profile,
                    )

                data_dic["profile"] = profile
            elif (
                self.config.dryrun_profilename
                and self.config.dryrun_profilename == payload["profile"]
            ):
                self.logger.debug(
                    "Order.add_profile_to_order(): adding dryrun profile to order"
                )
                data_dic["profile"] = payload["profile"]
            else:
                self.logger.warning(
                    "Ignore submitted profile '%s' as no profiles are configured.",
                    payload["profile"],
                )
        self.logger.debug("Order.add_profile_to_order() ended with %s", error)
        return error, data_dic

    def _apply_eab_profile(self, account_name: str) -> None:
        """Apply EAB profile settings to the order configuration."""
        self.logger.debug(
            "Order._apply_eab_profile() - apply eab profile setting for account %s",
            account_name,
        )

        if not self.config.eab_profiling:
            return

        try:
            account_dic = self.repository.account_lookup("name", account_name)
        except Exception as err_:
            self.logger.critical(
                "Database error: failed to look up account list: %s", err_
            )
            account_dic = {}

        eab_kid = account_dic.get("eab_kid") if account_dic else None

        if not eab_kid:
            return

        try:
            with self.config.eab_handler(self.logger) as eab_handler:
                profile_dic = eab_handler.key_file_load()
                # load allowed_iplist and allowed_domainlist from eab profile if specified
                self.config.allowed_iplist = self._load_eab_profile_param(
                    profile_dic, eab_kid, "allowed_iplist", self.config.allowed_iplist
                )
                self.config.allowed_domainlist = self._load_eab_profile_param(
                    profile_dic,
                    eab_kid,
                    "allowed_domainlist",
                    self.config.allowed_domainlist,
                )
                self.config.wildcard_certificate_disable = self._load_eab_profile_param(
                    profile_dic,
                    eab_kid,
                    "wildcard_certificate_disable",
                    self.config.wildcard_certificate_disable,
                )
                self.config.ca_error_details_forward = self._load_eab_profile_param(
                    profile_dic,
                    eab_kid,
                    "ca_error_details_forward",
                    self.config.ca_error_details_forward,
                    section="cahandler",
                )

                eab_profile_dic = self._load_eab_profile_mapping(profile_dic, eab_kid)
                self._apply_eab_profile_mapping(account_name, eab_profile_dic)

        except Exception as err:
            self.logger.error(
                "Failed to process EAB profile for Account %s (kid: %s): %s",
                account_name,
                eab_kid,
                err,
            )

    def _load_eab_profile_mapping(self, profile_dic, eab_kid) -> Dict[str, bool]:
        """Load and normalize profile mapping from EAB profile."""
        if not self.config.profile_mapping_field:
            return {}

        eab_profile = self._load_eab_profile_param(
            profile_dic,
            eab_kid,
            self.config.profile_mapping_field,
            None,
        )
        return self._profile_mapping_to_dict(eab_profile)

    def _apply_eab_profile_mapping(
        self, account_name: str, eab_profile_dic: Dict[str, bool]
    ) -> None:
        """Overwrite configured profiles with EAB profile mapping if provided."""
        if eab_profile_dic:
            self.logger.debug(
                "Order._apply_eab_profile() - overwrite profile information from eab profile for account %s",
                account_name,
            )
            self.config.profiles = eab_profile_dic
            if self.config.profiles_check_disable:
                self.logger.debug(
                    "Order._apply_eab_profile() - enabling profile validation because EAB profile mapping is present"
                )
                self.config.profiles_check_disable = False

    def _profile_mapping_to_dict(self, profile_value) -> Dict[str, bool]:
        """Normalize profile mapping value to dict keys."""
        if profile_value is None:
            return {}

        if isinstance(profile_value, str):
            key = profile_value.strip()
            return {key: True} if key else {}

        if isinstance(profile_value, list):
            return {
                str(item).strip(): True
                for item in profile_value
                if item is not None and str(item).strip()
            }

        if isinstance(profile_value, dict):
            # already in target format
            return profile_value

        self.logger.warning(
            "Unsupported profile mapping type: %s",
            type(profile_value).__name__,
        )
        return {}

    def create_order(
        self, payload: Dict[str, str], account_name: str
    ) -> Tuple[str, str, str, Dict[str, str], str]:
        """Create a new order and add it to the database."""
        self.logger.debug("Order.create_order(%s)", account_name)

        error = None
        detail = None
        auth_dic = {}
        order_name = generate_random_string(self.logger, 12)
        expires = uts_now() + self.config.validity

        # apply eab profiling if enabled
        if self.config.eab_profiling and self.config.eab_handler:
            self._apply_eab_profile(account_name)

        if "identifiers" in payload:
            data_dic = {"status": 2, "expires": expires, "account": account_name}
            data_dic["name"] = order_name
            data_dic["identifiers"] = json.dumps(payload["identifiers"])
            error, detail = self._check_identifiers_validity(payload["identifiers"])
            if error:
                data_dic["status"] = 1
            else:
                if "profile" in payload:
                    error, data_dic = self.add_profile_to_order(data_dic, payload)
                    if error == self.error_msg_dic["invalidprofile"]:
                        detail = "Invalid profile specified"
            error = self._add_order_and_authorizations(
                data_dic, auth_dic, payload, error
            )
        else:
            error = self.error_msg_dic["unsupportedidentifier"]

        self.logger.debug("Order.create_order() ended")
        return (error, detail, order_name, auth_dic, uts_to_date_utc(expires))

    def _load_eab_profile_param(
        self, profile_dic, eab_kid, param_name, default=None, section="order"
    ):
        """Helper to load allowed_iplist or allowed_domainlist from EAB profile."""
        profile_entry = profile_dic.get(eab_kid, {})
        order_cfg = profile_entry.get(section, {})

        # Prefer order section and honor explicit falsy values like False or []
        if param_name in order_cfg:
            value = order_cfg.get(param_name)
            self.logger.debug(
                "Order._load_eab_profile_param() - apply %s from eab profile.",
                param_name,
            )
            return value

        # Backward compatibility for deprecated cahandler section
        if section == "order":
            cahandler_cfg = profile_entry.get("cahandler", {})
            if param_name in cahandler_cfg:
                value = cahandler_cfg.get(param_name)
                if param_name != self.config.profile_mapping_field:
                    self.logger.warning(
                        "%s parameter found in cahandler section of the eab-profile - this is deprecated, please use the order section",
                        param_name,
                    )
                self.logger.debug(
                    "Order._load_eab_profile_param() - apply %s from eab profile.",
                    param_name,
                )
                return value

        return default

    def _load_header_info_config(self, config_dic: Dict[str, str]):
        """Load header info list from config file."""
        self.logger.debug("Order._load_header_info_config()")
        if "Order" in config_dic and "header_info_list" in config_dic["Order"]:
            try:
                self.config.header_info_list = json.loads(
                    config_dic["Order"]["header_info_list"]
                )
            except Exception as err_:
                self.logger.warning(
                    "Failed to parse header_info_list from configuration: %s",
                    err_,
                )
        self.logger.debug("Order._load_header_info_config() ended")

    def _load_order_config(self, config_dic: Dict[str, str]):
        """Load order-related configuration from file."""
        self.logger.debug("Order._load_order_config()")
        if "Challenge" in config_dic:
            self.config.sectigo_sim = config_dic.getboolean(
                "Challenge", "sectigo_sim", fallback=False
            )
        if "Order" in config_dic:
            self.config.tnauthlist_support = config_dic.getboolean(
                "Order", "tnauthlist_support", fallback=False
            )
            self.config.email_identifier_support = config_dic.getboolean(
                "Order", "email_identifier_support", fallback=False
            )
            self.config.email_identifier_rewrite = config_dic.getboolean(
                "Order", "email_identifier_rewrite", fallback=False
            )
            self.config.expiry_check_disable = config_dic.getboolean(
                "Order", "expiry_check_disable", fallback=False
            )
            self.config.idempotent_finalize = config_dic.getboolean(
                "Order", "idempotent_finalize", fallback=False
            )
            self.config.wildcard_certificate_disable = config_dic.getboolean(
                "Order", "wildcard_certificate_disable", fallback=False
            )
            self.config.ca_error_details_forward = config_dic.getboolean(
                "CAhandler", "ca_error_details_forward", fallback=False
            )

            try:
                self.config.retry_after = int(
                    config_dic.get(
                        "Order", "retry_after_timeout", fallback=self.config.retry_after
                    )
                )
            except Exception:
                self.logger.warning(
                    "Failed to parse retry_after from configuration: %s",
                    config_dic["Order"].get("retry_after_timeout", None),
                )
            try:
                self.config.validity = int(
                    config_dic.get("Order", "validity", fallback=self.config.validity)
                )
            except Exception:
                self.logger.warning(
                    "Failed to parse validity from configuration: %s",
                    config_dic["Order"].get("validity", None),
                )
            try:
                self.config.identifier_limit = int(
                    config_dic.get("Order", "identifier_limit", fallback=20)
                )
            except Exception:
                self.logger.warning(
                    "Failed to parse identifier_limit from configuration: %s",
                    config_dic["Order"].get("identifier_limit", None),
                )
            _not_used, self.config.dryrun_profilename = config_dryrun_load(
                self.logger, config_dic
            )

        self.logger.debug("Order._load_order_config() ended")

    def _load_profile_config(self, config_dic: Dict[str, str]):
        """Load profiles from file or database."""
        self.logger.debug("Order._load_profile_config()")
        self._load_profiles_from_config(config_dic)
        self._load_profiles_from_db_if_sync(config_dic)
        self._maybe_disable_profile_check(config_dic)
        self.logger.debug("Order._load_profile_config() ended")

    def _load_profiles_from_config(self, config_dic: Dict[str, str]):
        """Load profiles from configuration file."""
        if "Order" in config_dic and "profiles" in config_dic["Order"]:
            self.logger.debug("Order._config_load(): profile check enabled")
            self.config.profiles_check_disable = False
            self.config.profiles = config_profile_load(self.logger, config_dic)

    def _load_profiles_from_db_if_sync(self, config_dic: Dict[str, str]):
        """Load profiles from database if profiles_sync is set."""
        if "CAhandler" in config_dic and "profiles_sync" in config_dic["CAhandler"]:
            self.config.profiles_sync = config_dic.getboolean(
                "CAhandler", "profiles_sync", fallback=False
            )
            if self.config.profiles_sync:
                self.logger.debug(
                    "Order._config_load(): profile_sync set. Loading profiles"
                )
                try:
                    profiles = self.repository.hkparameter_get("profiles")
                except Exception as err:
                    self.logger.critical(
                        "Database error: failed to get profile list: %s", err
                    )
                    profiles = None
                if profiles:
                    self._set_profiles_from_db(profiles)

    def _set_profiles_from_db(self, profiles):
        """Set profiles from database string."""
        try:
            profile_dic = json.loads(profiles)
            self.config.profiles = profile_dic.get("profiles", {})
        except Exception as err_:
            self.logger.error(
                "Error when loading the profiles parameter from database: %s", err_
            )

    def _maybe_disable_profile_check(self, config_dic: Dict[str, str]):
        """Disable profile check"""
        if self.config.profiles and "Order" in config_dic:
            self.config.profiles_check_disable = config_dic.getboolean(
                "Order", "profiles_check_disable", fallback=False
            )

    def _load_configuration(self):
        """Load all configuration from file."""
        self.logger.debug("Order._load_configuration()")
        config_dic = load_config()
        # load order config
        self._load_order_config(config_dic)
        self._load_header_info_config(config_dic)
        if "Authorization" in config_dic:
            try:
                self.config.authz_validity = int(
                    config_dic.get(
                        "Authorization", "validity", fallback=self.config.authz_validity
                    )
                )
            except Exception:
                self.logger.warning(
                    "Failed to parse authz validity from configuration: %s",
                    config_dic["Authorization"].get("validity", None),
                )
        if "Directory" in config_dic and "url_prefix" in config_dic["Directory"]:
            self.path_dic = {
                k: config_dic["Directory"]["url_prefix"] + v
                for k, v in self.path_dic.items()
            }
        self._load_profile_config(config_dic)

        # load allowed domainlist
        self.config.allowed_domainlist = config_allowed_domainlist_load(
            self.logger, config_dic
        )
        # load allowed iplist
        self.config.allowed_iplist = config_allowed_iplist_load(self.logger, config_dic)
        # load profiling
        (
            self.config.eab_profiling,
            self.config.eab_handler,
        ) = config_eab_profile_load(self.logger, config_dic)

        self.config.profile_mapping_field = self._load_profile_mapping_field(config_dic)

        self.logger.debug("Order._load_configuration() ended.")

    def _load_profile_mapping_field(self, config_dic: Dict[str, str]) -> Optional[str]:
        """Load profile mapping field from configuration."""
        self.logger.debug("Order._load_profile_mapping_field()")
        ca_handler_module = ca_handler_load(self.logger, config_dic)

        if not ca_handler_module:
            return None

        # Backward compatibility: allow module-level definition.
        if hasattr(ca_handler_module, "profile_mapping_field"):
            mapping_field = getattr(ca_handler_module, "profile_mapping_field")
            self.logger.debug(
                "Order._load_profile_mapping_field() - profile_mapping_field '%s' loaded from CA handler module",
                mapping_field,
            )
            return mapping_field

        ca_handler_class = getattr(ca_handler_module, "CAhandler", None)
        if not ca_handler_class:
            return None

        # Best effort: try instance attributes, but never fail startup on constructor issues.
        ca_handler_obj = None
        try:
            ca_handler_obj = ca_handler_class(logger=self.logger)
        except TypeError:
            try:
                ca_handler_obj = ca_handler_class()
            except Exception as err_:
                self.logger.warning(
                    "Order._load_profile_mapping_field() - failed to instantiate CAhandler without logger: %s",
                    err_,
                )
        except Exception as err_:
            self.logger.warning(
                "Order._load_profile_mapping_field() - failed to instantiate CAhandler with logger: %s",
                err_,
            )

        if ca_handler_obj:
            mapping_field = getattr(ca_handler_obj, "profile_mapping_field", None)
            if mapping_field:
                self.logger.debug(
                    "Order._load_profile_mapping_field() - profile_mapping_field '%s' loaded from CAhandler instance",
                    mapping_field,
                )
                return mapping_field

        # Check for class-level attribute if instance attribute is not found
        mapping_field = getattr(ca_handler_class, "profile_mapping_field", None)
        if mapping_field:
            self.logger.debug(
                "Order._load_profile_mapping_field() - profile_mapping_field '%s' loaded from CAhandler class",
                mapping_field,
            )
            return mapping_field

        return None

    def _name_get(self, url: str) -> str:
        """get ordername"""
        self.logger.debug("Order._name_get(%s)", url)

        url_dic = parse_url(self.logger, url)
        order_name = url_dic["path"].replace(self.path_dic["order_path"], "")
        if "/" in order_name:
            order_name, _sinin = order_name.split("/", 1)
        self.logger.debug("Order._name_get() ended")
        return order_name

    def are_identifiers_allowed(self, identifiers_list: List[str]) -> Tuple[str, str]:
        """Check if the provided identifiers are allowed."""
        self.logger.debug("Order.are_identifiers_allowed()")
        error = None
        detail = None
        allowed_identifiers = self._get_allowed_identifier_types()
        for identifier in identifiers_list:
            error, detail = self._check_single_identifier(
                identifier, allowed_identifiers
            )
            if error:
                break
        self.logger.debug("Order.are_identifiers_allowed() ended with: %s", error)
        return error, detail

    def _get_allowed_identifier_types(self) -> List[str]:
        allowed = ["dns", "ip"]
        if self.config.tnauthlist_support:
            allowed.append("tnauthlist")
        if self.config.email_identifier_support:
            allowed.append("email")
        return allowed

    def _check_single_identifier(
        self, identifier: dict, allowed_identifiers: List[str]
    ) -> Tuple[str, str]:
        """Check if a single identifier is allowed."""
        self.logger.debug("Order._check_single_identifier(%s)", identifier)

        error = self._validate_identifier_structure(identifier)
        if error:
            return error

        id_type = identifier["type"].lower()

        error = self._validate_identifier_type_supported(
            identifier["type"], id_type, allowed_identifiers
        )
        if error:
            return error

        error = self._validate_identifier_value(identifier, id_type)
        if error:
            return error

        if id_type == "dns":
            return self._validate_dns_identifier_policy(identifier)

        if id_type == "ip":
            return self._validate_ip_identifier_policy(identifier)

        return None, None

    def _validate_identifier_structure(
        self, identifier: dict
    ) -> Optional[Tuple[str, str]]:
        """Validate mandatory identifier fields are present."""
        if "type" not in identifier:
            self.logger.error("Identifier type is missing")
            return self.error_msg_dic["malformed"], "Identifier type is missing"

        if "value" not in identifier:
            self.logger.error("Identifier value is missing")
            return self.error_msg_dic["malformed"], "Identifier value is missing"

        return None

    def _validate_identifier_type_supported(
        self, identifier_type: str, id_type: str, allowed_identifiers: List[str]
    ) -> Optional[Tuple[str, str]]:
        """Validate identifier type is supported."""
        if id_type in allowed_identifiers:
            return None

        self.logger.error("Identifier type %s not supported", identifier_type)
        return (
            self.error_msg_dic["unsupportedidentifier"],
            f"Identifier type {identifier_type} not supported",
        )

    def _validate_identifier_value(
        self, identifier: dict, id_type: str
    ) -> Optional[Tuple[str, str]]:
        """Validate identifier value format."""
        if validate_identifier(
            self.logger,
            id_type,
            identifier["value"],
            self.config.tnauthlist_support,
        ):
            return None

        self.logger.error(
            "Identifier value %s not allowed for type %s",
            identifier["value"],
            identifier["type"],
        )
        return (
            self.error_msg_dic["rejectedidentifier"],
            f'identifier value {identifier["value"]} not allowed',
        )

    def _validate_dns_identifier_policy(
        self, identifier: dict
    ) -> Tuple[Optional[str], Optional[str]]:
        """Validate DNS-specific policy constraints."""
        if self.config.wildcard_certificate_disable and identifier["value"].startswith(
            "*."
        ):
            self.logger.error(
                "Wildcard identifier %s not allowed by configuration",
                identifier["value"],
            )
            return (
                self.error_msg_dic["rejectedidentifier"],
                f'Wildcard identifier {identifier["value"]} not allowed',
            )

        if not self.config.allowed_domainlist:
            return None, None

        domain_allowed = self._is_domain_allowed_for_identifier(identifier)
        if domain_allowed:
            return None, None

        self.logger.error(
            "FQDN/SAN %s not allowed by configuration",
            identifier["value"],
        )
        return (
            self.error_msg_dic["rejectedidentifier"],
            f'FQDN/SAN {identifier["value"]} not allowed by configuration',
        )

    def _is_domain_allowed_for_identifier(self, identifier: dict) -> bool:
        """Check allowlist for DNS identifiers with wildcard fallback handling."""
        domain_value = identifier["value"]
        wildcard_requested = bool(identifier.get("wildcard"))

        self.logger.debug(
            "Order._check_single_identifier() - Evaluating allowed_domainlist for value='%s' (wildcard flag: %s)",
            domain_value,
            wildcard_requested,
        )

        domain_allowed = is_domain_whitelisted(
            self.logger,
            domain_value,
            self.config.allowed_domainlist,
        )

        self.logger.debug(
            "Order._check_single_identifier() - allowed_domainlist check result for value='%s': %s",
            domain_value,
            domain_allowed,
        )

        if (
            not domain_allowed
            and wildcard_requested
            and domain_value
            and not domain_value.startswith("*.")
        ):
            wildcard_value = f"*.{domain_value}"
            self.logger.debug(
                "Order._check_single_identifier() - Reconstructed wildcard value for allowed_domainlist check: '%s' -> '%s'",
                domain_value,
                wildcard_value,
            )
            domain_allowed = is_domain_whitelisted(
                self.logger,
                wildcard_value,
                self.config.allowed_domainlist,
            )
            self.logger.debug(
                "Order._check_single_identifier() - allowed_domainlist check result for reconstructed wildcard value='%s': %s",
                wildcard_value,
                domain_allowed,
            )
        else:
            self.logger.debug(
                "Order._check_single_identifier() - Skipping reconstructed wildcard allowed_domainlist check (domain_allowed=%s, wildcard=%s, value_present=%s, value_has_wildcard_prefix=%s)",
                domain_allowed,
                wildcard_requested,
                bool(domain_value),
                bool(domain_value and domain_value.startswith("*.")),
            )

        return domain_allowed

    def _validate_ip_identifier_policy(
        self, identifier: dict
    ) -> Tuple[Optional[str], Optional[str]]:
        """Validate IP-specific policy constraints."""
        if not self.config.allowed_iplist:
            return None, None

        if is_ip_whitelisted(
            self.logger,
            identifier["value"],
            self.config.allowed_iplist,
        ):
            return None, None

        self.logger.error(
            "IP address %s not allowed by configuration",
            identifier["value"],
        )
        return (
            self.error_msg_dic["rejectedidentifier"],
            f'IP address {identifier["value"]} not allowed by configuration',
        )

    def _rewrite_email_identifiers(
        self, identifiers_list: List[Dict[str, str]]
    ) -> List[Dict[str, str]]:
        """Rewrite DNS identifiers with @ to email identifiers."""
        self.logger.debug("Order._rewrite_email_identifiers()")

        if (
            self.config.email_identifier_support
            and self.config.email_identifier_rewrite
        ):
            identifiers_modified = []
            for ident in identifiers_list:
                if (
                    "type" in ident
                    and "value" in ident
                    and ident["type"].lower() == "dns"
                    and "@" in ident["value"]
                ):
                    self.logger.info(
                        "Rewrite DNS identifier '%s' to email identifier",
                        ident["value"],
                    )
                    ident["type"] = "email"
                identifiers_modified.append(ident)
        else:
            identifiers_modified = identifiers_list

        self.logger.debug("Order._rewrite_email_identifiers() ended")
        return identifiers_modified

    def _check_identifier_limit(self, identifiers_list: List[str]) -> bool:
        """Check and log if identifier limit is exceeded."""
        self.logger.debug("Order._check_identifier_limit()")
        error = False
        if len(identifiers_list) > self.config.identifier_limit:
            self.logger.warning(
                "Number of identifiers %d exceeds limit %d",
                len(identifiers_list),
                self.config.identifier_limit,
            )
            error = True
        return error

    def _check_identifiers_validity(
        self, identifiers_list: List[str]
    ) -> Tuple[str, str]:
        """Check validity of identifiers in the order."""
        self.logger.debug("Order._check_identifiers_validity(%s)", identifiers_list)
        # make a deep copy to avoid modifying the original list
        identifiers_list = copy.deepcopy(identifiers_list)

        if identifiers_list and isinstance(identifiers_list, list):

            # rewrite email identifiers if configured
            identifiers_list = self._rewrite_email_identifiers(identifiers_list)

            # check identifier limit
            if self._check_identifier_limit(identifiers_list):
                return (
                    self.error_msg_dic["rejectedidentifier"],
                    "identifier limit exceeded",
                )

            # check if identifier types and values are allowed
            error, detail = self.are_identifiers_allowed(identifiers_list)
            if error:
                self.logger.debug(
                    "Order._check_identifiers_validity() ended with %s:", error
                )
                return error, detail

        else:
            # malformed identifiers list
            error = self.error_msg_dic["malformed"]
            detail = "malformed identifiers list"

        self.logger.debug("Order._check_identifiers_validity() done with %s:", error)
        return error, detail

    def _get_order_info(self, order_name: str) -> Dict[str, str]:
        """List details of an order. Returns order dict or empty dict on error."""
        self.logger.debug("Order._get_order_info(%s)", order_name)
        try:
            result = self.repository.order_lookup("name", order_name)
        except Exception as err_:
            self.logger.critical("Database error: failed to look up order: %s", err_)
            result = None
        return result

    def _get_order_account_name(self, order_name: str) -> Optional[str]:
        """Lookup account name for an order (needed for EAB profile at finalize)."""
        self.logger.debug("Order._get_order_account_name(%s)", order_name)
        try:
            order_dic = self.repository.order_lookup(
                "name", order_name, vlist=["account__name"]
            )
        except Exception as err_:
            self.logger.critical(
                "Database error: failed to look up order account: %s", err_
            )
            return None
        if not order_dic:
            return None
        return order_dic.get("account__name") or order_dic.get("account")

    def _header_info_lookup(self, header: Optional[Dict[str, Any]]) -> str:
        """lookup header information and serialize them in a string"""
        self.logger.debug("Order._header_info_lookup()")

        header_info_dic = {}
        if header and self.config.header_info_list:
            for ele in self.config.header_info_list:
                if ele in header:
                    header_info_dic[ele] = header[ele]

        result = None
        if header_info_dic:
            result = json.dumps(header_info_dic)

        self.logger.debug(
            "Order._header_info_lookup() ended with: %s keys in dic",
            len(header_info_dic.keys()),
        )
        return result

    def _finalize_csr(
        self, order_name: str, payload: Dict[str, str], header: str = None
    ) -> Tuple[int, str, str, str]:
        """Handle CSR finalization for an order"""
        self.logger.debug("Order._finalize_csr(%s)", order_name)

        message = None
        # lookup header information
        header_info = self._header_info_lookup(header)
        # this is a new request
        code, certificate_name, detail = self._process_csr(
            order_name, payload["csr"], header_info
        )

        # change status only if we do not have a poll_identifier (stored in detail variable)
        if code == 200:
            if not detail:
                # update order_status / set to valid
                self.repository.order_update({"name": order_name, "status": "valid"})
        elif certificate_name == "timeout":
            code = 200
            message = certificate_name
        elif certificate_name == "urn:ietf:params:acme:error:rejectedIdentifier":
            code = 401
            message = certificate_name
        elif detail == DRYRUN_ENROLLMENT_SKIPPED_DETAIL:
            message = "urn:ietf:params:acme:error:unauthorized"
        else:
            message = certificate_name
            if not detail:
                detail = ENROLLMENT_FAILED_DETAIL

        self.logger.debug("Order._finalize_csr() ended")
        return (code, message, detail, certificate_name)

    def _finalize_order(
        self, order_name: str, payload: Dict[str, str], header: str = None
    ) -> Tuple[int, str, str, str]:
        """finalize request"""
        self.logger.debug("Order._finalize_order()")

        certificate_name = None
        message = None
        detail = None

        # lookup order-status (must be ready to proceed)
        order_dic = self._get_order_info(order_name)
        if "status" in order_dic and order_dic["status"] == "ready":
            # update order_status / set to processing
            self.repository.order_update({"name": order_name, "status": "processing"})
            if "csr" in payload:
                code, message, detail, certificate_name = self._finalize_csr(
                    order_name, payload, header
                )
            else:
                code = 400
                message = self.error_msg_dic["badcsr"]
                detail = "csr is missing in payload"
        elif (
            "status" in order_dic
            and order_dic["status"] == "valid"
            and self.config.idempotent_finalize
        ):
            self.logger.debug(
                "Order._finalize_order(): kind of polling request - order is already valid - lookup certificate"
            )
            code = 200
            try:
                cert_dic = self.repository.certificate_lookup("order__name", order_name)
            except Exception as err_:
                self.logger.critical(
                    "Database error: Certificate lookup failed: %s", err_
                )
                cert_dic = {}
            if cert_dic and "name" in cert_dic:
                certificate_name = cert_dic["name"]
        else:
            code = 403
            message = self.error_msg_dic["ordernotready"]
            detail = "Order is not ready"

        self.logger.debug("Order._finalize_order() ended")
        return (code, message, detail, certificate_name)

    def _process_order_request(
        self,
        order_name: str,
        protected: Dict[str, str],
        payload: Dict[str, str],
        header: Optional[str] = None,
    ) -> Tuple[int, str, str, str]:
        """process order"""
        self.logger.debug("Order._process_order_request({%s)", order_name)

        certificate_name = None
        message = None
        detail = None

        if "url" in protected:
            if "finalize" in protected["url"]:
                code, message, detail, certificate_name = self._finalize_order(
                    order_name, payload, header
                )
            else:
                self.logger.debug("polling request()")
                code = 200
                try:
                    cert_dic = self.repository.certificate_lookup(
                        "order__name", order_name
                    )
                except Exception as err_:
                    self.logger.critical(
                        "Database error: Certificate lookup failed: %s", err_
                    )
                    cert_dic = {}
                if cert_dic and "name" in cert_dic:
                    certificate_name = cert_dic["name"]
        else:
            code = 400
            message = self.error_msg_dic["malformed"]
            detail = "url is missing in protected"

        self.logger.debug(
            "Order._process_order_request() ended with order:%s %s:%s:%s",
            order_name,
            code,
            message,
            detail,
        )
        return (code, message, detail, certificate_name)

    def _apply_finalize_eab_profile(self, order_name: str) -> None:
        """Apply EAB profile for finalize flow when enabled."""
        if not (self.config.eab_profiling and self.config.eab_handler):
            return

        account_name = self._get_order_account_name(order_name)
        if account_name:
            self._apply_eab_profile(account_name)

    def _map_csr_enrollment_result(
        self, certificate_name: str, error: str, detail: str
    ) -> Tuple[int, str, str]:
        """Map CA enrollment result to ACME response tuple."""
        if (
            error == "urn:ietf:params:acme:error:rejectedIdentifier"
            or detail == DRYRUN_ENROLLMENT_SKIPPED_DETAIL
        ):
            return 401, error, detail

        if not error:
            return 200, certificate_name, detail

        if error == self.error_msg_dic["serverinternal"]:
            return 500, error, detail

        return 400, error, detail

    def _enroll_certificate_for_order(
        self, order_name: str, csr: str, header_info: str
    ) -> Tuple[int, str, str]:
        """Store CSR and perform enrollment for an order."""
        with Certificate(self.debug, self.server_name, self.logger) as certificate:
            certificate.config.ca_error_details_forward = (
                self.config.ca_error_details_forward
            )
            certificate_name = certificate.store_csr(order_name, csr, header_info)
            if not certificate_name:
                return (
                    500,
                    self.error_msg_dic["serverinternal"],
                    "CSR processing failed",
                )

            error, detail = certificate.enroll_and_store(
                certificate_name, csr, order_name
            )
            return self._map_csr_enrollment_result(certificate_name, error, detail)

    def _process_csr(
        self, order_name: str, csr: str, header_info: str
    ) -> Tuple[int, str, str]:
        """process certificate signing request"""
        self.logger.debug("Order._process_csr(%s)", order_name)

        order_dic = self._get_order_info(order_name)
        if not order_dic:
            code = 400
            message = self.error_msg_dic["unauthorized"]
            detail = f"order: {order_name} not found"
        else:
            self._apply_finalize_eab_profile(order_name)
            csr = b64_url_recode(self.logger, csr)
            code, message, detail = self._enroll_certificate_for_order(
                order_name, csr, header_info
            )

        self.logger.debug(
            "Order._process_csr() ended with order:%s %s:{%s:%s",
            order_name,
            code,
            message,
            detail,
        )
        return (code, message, detail)

    def _order_dic_create(self, tmp_dic: Dict[str, str]) -> Dict[str, str]:
        """create order dictionary"""
        self.logger.debug("Order._order_dic_create()")

        order_dic = {}
        if "status" in tmp_dic:
            order_dic["status"] = tmp_dic["status"]
        if "expires" in tmp_dic:
            order_dic["expires"] = uts_to_date_utc(tmp_dic["expires"])
        if "notbefore" in tmp_dic and tmp_dic["notbefore"] != 0:
            order_dic["notBefore"] = uts_to_date_utc(tmp_dic["notbefore"])
        if "notafter" in tmp_dic and tmp_dic["notafter"] != 0:
            order_dic["notAfter"] = uts_to_date_utc(tmp_dic["notafter"])
        if "identifiers" in tmp_dic:
            try:
                order_dic["identifiers"] = json.loads(tmp_dic["identifiers"])
            except Exception:
                self.logger.error(
                    "Error while parsing the identifier %s",
                    tmp_dic["identifiers"],
                )

        self.logger.debug("Order._order_dic_create() ended")
        return order_dic

    def _get_authorization_list(self, order_name: str) -> List[str]:
        """Lookup authorization list. Returns list or empty list on error."""
        self.logger.debug("Order._get_authorization_list(%s)", order_name)
        try:
            authz_list = self.repository.authorization_lookup(
                "order__name", order_name, ["name", "status__name"]
            )
        except Exception as err_:
            self.logger.critical(
                "Database error: failed to look up authorization list: %s", err_
            )
            authz_list = []
        self.logger.debug("Order._get_authorization_list() ended")
        return authz_list

    def _update_validity_list(
        self, authz_list: List[str], order_dic: Dict[str, str], order_name: str
    ):
        """update validity list and order status"""
        self.logger.debug("Order._update_validity_list()")
        validity_list = []
        for authz in authz_list:
            if "name" in authz:
                order_dic["authorizations"].append(
                    f'{self.server_name}{self.path_dic["authz_path"]}{authz["name"]}'
                )
            if "status__name" in authz:
                if authz["status__name"] == "valid":
                    validity_list.append(True)
                else:
                    validity_list.append(False)

        # update orders status from pending to ready
        if validity_list and "status" in order_dic:
            if False not in validity_list and order_dic["status"] == "pending":
                self.repository.order_update({"name": order_name, "status": "ready"})

        self.logger.debug("Order.get_order_details() ended")

    def get_order_details(self, order_name: str) -> Dict[str, str]:
        """Show order details based on order name."""
        self.logger.debug("Order.get_order_details(%s)", order_name)

        order_dic = {}
        tmp_dic = self._get_order_info(order_name)
        if tmp_dic:
            # create order dictionary and lookup authorization list
            order_dic = self._order_dic_create(tmp_dic)
            authz_list = self._get_authorization_list(order_name)
            if authz_list:
                order_dic["authorizations"] = []
                # collect status of different authorizations in list and update order status
                self._update_validity_list(authz_list, order_dic, order_name)
        self.logger.debug("Order.get_order_details() ended")
        return order_dic

    def invalidate_expired_orders(
        self, timestamp: int = None
    ) -> Tuple[List[str], List[str]]:
        """Invalidate orders that have expired."""
        self.logger.debug("Order.invalidate_expired_orders(%s)", timestamp)

        if not timestamp:
            timestamp = uts_now()
            self.logger.debug(
                "Order.invalidate_expired_orders(): set timestamp to %s", timestamp
            )

        field_list = [
            "id",
            "name",
            "expires",
            "identifiers",
            "created_at",
            "status__id",
            "status__name",
            "account__id",
            "account__name",
            "account__contact",
        ]
        try:
            order_list = self.repository.orders_invalid_search(
                "expires", timestamp, vlist=field_list, operant="<="
            )
        except Exception as err_:
            self.logger.critical(
                "Database error: failed to search for expired orders: %s", err_
            )
            order_list = []
        output_list = []
        for order in order_list:
            # select all orders which are not invalid
            if (
                "name" in order
                and "status__name" in order
                and order["status__name"] != "invalid"
            ):
                # change status and add to output list
                output_list.append(order)
                data_dic = {"name": order["name"], "status": "invalid"}
                try:
                    self.repository.order_update(data_dic)
                except Exception as err_:
                    self.logger.critical(
                        "Database error: failed to update order status to invalid: %s",
                        err_,
                    )

        self.logger.debug(
            "Order.invalidate_expired_orders() ended: %s orders identified",
            len(output_list),
        )
        return (field_list, output_list)

    def create_from_content(self, content: str) -> Dict[str, str]:
        """new order request (renamed from new)"""
        self.logger.debug("Order.create_from_content()")

        response_dic = {}
        # check message
        code, message, detail, _protected, payload, account_name = self.message.check(
            content
        )

        if code == 200:
            error, detail, order_name, auth_dic, expires = self.create_order(
                payload, account_name
            )
            if not error:
                code = 201
                response_dic["header"] = {}
                response_dic["header"][
                    "Location"
                ] = f'{self.server_name}{self.path_dic["order_path"]}{order_name}'
                response_dic["data"] = {}
                response_dic["data"]["identifiers"] = []
                response_dic["data"]["authorizations"] = []
                response_dic["data"]["status"] = "pending"
                response_dic["data"]["expires"] = expires
                response_dic["data"][
                    "finalize"
                ] = f'{self.server_name}{self.path_dic["order_path"]}{order_name}/finalize'
                for auth_name, value in auth_dic.items():
                    response_dic["data"]["authorizations"].append(
                        f'{self.server_name}{self.path_dic["authz_path"]}{auth_name}'
                    )
                    response_dic["data"]["identifiers"].append(value)
            elif error in [
                self.error_msg_dic["rejectedidentifier"],
                self.error_msg_dic["invalidprofile"],
            ]:
                code = 403
                message = error
                if not detail:
                    detail = "Some of the requested identifiers got rejected"
            elif error == self.error_msg_dic["malformed"]:
                code = 400
                message = error
                if not detail:
                    detail = "One of the requested identifiers is not supported"
            else:
                code = 400
                message = error
                detail = "Could not process order"

        # prepare/enrich response
        status_dic = {"code": code, "type": message, "detail": detail}
        response_dic = self.message.prepare_response(response_dic, status_dic)

        self.logger.debug(
            "Order.create_from_content() returns: %s", json.dumps(response_dic)
        )
        return response_dic

    def _parse_order_message(
        self, protected: Dict[str, str], payload: Dict[str, str], header: str = None
    ) -> Tuple[int, str, str, str, str]:
        """parse new order message"""
        self.logger.debug("Order._parse_order_message()")

        order_name = certificate_name = None

        if "url" in protected:
            order_name = self._name_get(protected["url"])
            if order_name:
                order_dic = self.get_order_details(order_name)
                if order_dic:
                    (
                        code,
                        message,
                        detail,
                        certificate_name,
                    ) = self._process_order_request(
                        order_name, protected, payload, header
                    )
                else:
                    code = 403
                    message = self.error_msg_dic["ordernotready"]
                    detail = "order not found"
            else:
                code = 400
                message = self.error_msg_dic["malformed"]
                detail = "order name is missing"
        else:
            code = 400
            message = self.error_msg_dic["malformed"]
            detail = "url is missing in protected"

        self.logger.debug("Order._parse_order_message() ended with code: %s", code)
        return (code, message, detail, certificate_name, order_name)

    def parse_order_content(self, content: str, header: str = None) -> Dict[str, str]:
        """parse order request (renamed from parse)"""
        self.logger.debug("Order.parse_order_content()")

        # invalidate expired orders
        if not self.config.expiry_check_disable:
            self.invalidate_expired_orders()

        response_dic = {}
        # check message
        code, message, detail, protected, payload, _account_name = self.message.check(
            content
        )

        if code == 200:
            # parse message
            (
                code,
                message,
                detail,
                certificate_name,
                order_name,
            ) = self._parse_order_message(protected, payload, header)

            if code == 200:
                # create response
                response_dic["header"] = {}
                response_dic["header"][
                    "Location"
                ] = f'{self.server_name}{self.path_dic["order_path"]}{order_name}'
                response_dic["data"] = self.get_order_details(order_name)
                if (
                    "status" in response_dic["data"]
                    and response_dic["data"]["status"] == "processing"
                ):
                    # set retry header as cert issuane is not completed.
                    response_dic["header"]["Retry-After"] = f"{self.config.retry_after}"
                response_dic["data"][
                    "finalize"
                ] = f'{self.server_name}{self.path_dic["order_path"]}{order_name}/finalize'
                # add the path to certificate if order-status is ready
                if (
                    certificate_name
                    and "status" in response_dic["data"]
                    and response_dic["data"]["status"] == "valid"
                ):
                    response_dic["data"][
                        "certificate"
                    ] = f'{self.server_name}{self.path_dic["cert_path"]}{certificate_name}'

        # prepare/enrich response
        status_dic = {"code": code, "type": message, "detail": detail}
        response_dic = self.message.prepare_response(response_dic, status_dic)

        self.logger.debug(
            "Order.parse_order_content() returns: %s", json.dumps(response_dic)
        )
        return response_dic

    # === Legacy API Compatibility ===

    def invalidate(self, timestamp: int = None) -> Tuple[List[str], List[str]]:
        """invalidate orders"""
        self.logger.debug(
            "Order.invalidate() - Compatibility wrapper for old method name"
        )
        return self.invalidate_expired_orders(timestamp)

    def new(self, content: str) -> Dict[str, str]:
        """new order request"""
        self.logger.debug("Order.new() - Compatibility wrapper for old method name")
        return self.create_from_content(content)

    def parse(self, content: str, header: str = None) -> Dict[str, str]:
        """parse order request"""
        self.logger.debug("Order.parse() - Compatibility wrapper for old method name")
        return self.parse_order_content(content, header)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""customized class for django json response"""

import json
from django.http import HttpResponse
from django.core.serializers.json import DjangoJSONEncoder


class JsonResponse(HttpResponse):
    """
    An HTTP response class that consumes data to be serialized to JSON.
    and changes the contentent type base do status code
    """

    def __init__(
        self,
        data,
        encoder=DjangoJSONEncoder,
        safe=True,
        json_dumps_params=None,
        **kwargs,
    ):
        if safe and not isinstance(data, dict):
            raise TypeError(
                "In order to allow non-dict objects to be serialized set the "
                "safe parameter to False."
            )
        if json_dumps_params is None:
            json_dumps_params = {"indent": 2}

        if "status" in kwargs and kwargs["status"] > 201:
            kwargs.setdefault("content_type", "application/problem+json")
        else:
            kwargs.setdefault("content_type", "application/json")

        data = json.dumps(data, cls=encoder, **json_dumps_params)
        super().__init__(content=data, **kwargs)



"""
TLS-ALPN-01 Challenge Validator.

Implements validation logic for TLS-ALPN-01 challenges according to RFC 8737.
"""

import json
from .base import ChallengeValidator, ChallengeContext, ValidationResult


class TlsAlpnChallengeValidator(ChallengeValidator):
    """Validator for TLS-ALPN-01 challenges."""

    def get_challenge_type(self) -> str:
        return "tls-alpn-01"

    def perform_validation(self, context: ChallengeContext) -> ValidationResult:
        """Perform TLS-ALPN-01 challenge validation."""
        self.logger.debug("TlsAlpnChallengeValidator.perform_validation()")
        try:
            from acme2certifier.acme_srv.helper import (
                fqdn_resolve,
                ip_validate,
                proxy_check,
                servercert_get,
                sha256_hash_hex,
                b64_encode,
            )
        except ImportError as e:
            return ValidationResult(
                success=False,
                invalid=True,
                error_message=f"Required dependencies not available: {e}",
                details={"import_error": str(e)},
            )

        # Determine SNI value
        if context.authorization_type == "dns":
            _, invalid, error_msg = fqdn_resolve(
                self.logger, context.authorization_value, context.dns_servers
            )
            if invalid:
                return ValidationResult(
                    success=False,
                    invalid=True,
                    error_message=json.dumps(
                        {
                            "status": 400,
                            "type": "urn:ietf:params:acme:error:dns",
                            "detail": (
                                f"DNS resolution failed: {error_msg}"
                                if error_msg
                                else "DNS resolution failed"
                            ),
                        }
                    ),
                )
            sni = context.authorization_value
        elif context.authorization_type == "ip":
            sni, invalid = ip_validate(self.logger, context.authorization_value)
            if invalid:
                return ValidationResult(
                    success=False,
                    invalid=True,
                    error_message=json.dumps(
                        {
                            "status": 400,
                            "type": "urn:ietf:params:acme:error:malformed",
                            "detail": f"Invalid IP address: {context.authorization_value}",
                        }
                    ),
                    details={"ip": context.authorization_value},
                )
        else:
            return ValidationResult(
                success=False,
                invalid=True,
                error_message=json.dumps(
                    {
                        "status": 400,
                        "type": "urn:ietf:params:acme:error:unsupported",
                        "detail": f"Unsupported authorization type: {context.authorization_type}",
                    }
                ),
                details={"type": context.authorization_type},
            )

        # Compute expected extension value
        sha256_digest = sha256_hash_hex(
            self.logger, f"{context.token}.{context.jwk_thumbprint}"
        )
        extension_value = b64_encode(
            self.logger, bytearray.fromhex(f"0420{sha256_digest}")
        )

        # Check for proxy configuration
        proxy_server = None
        if context.proxy_servers:
            proxy_server = proxy_check(
                self.logger, context.authorization_value, context.proxy_servers
            )

        # Get server certificate
        cert = servercert_get(
            self.logger, context.authorization_value, 443, proxy_server, sni
        )

        if not cert:
            return ValidationResult(
                success=False,
                invalid=False,
                error_message=json.dumps(
                    {
                        "status": 400,
                        "type": "urn:ietf:params:acme:error:incorrectResponse",
                        "detail": f"Unable to retrieve server certificate for {context.authorization_value}",
                    }
                ),
            )

        # Validate certificate extensions
        success = self._validate_certificate_extensions(
            cert, extension_value, context.authorization_value
        )

        self.logger.debug(
            "TlsAlpnChallengeValidator.perform_validation() ended with: %s", success
        )
        return ValidationResult(
            success=success,
            invalid=not success,
            error_message=(
                None
                if success
                else json.dumps(
                    {
                        "status": 403,
                        "type": "urn:ietf:params:acme:error:incorrectResponse",
                        "detail": "Certificate extension validation failed",
                    }
                )
            ),
            details={"expected_extension": extension_value, "sni": sni},
        )

    def _validate_certificate_extensions(
        self, cert: str, extension_value: str, fqdn: str
    ) -> bool:
        """Validate certificate extensions for TLS-ALPN challenge."""
        self.logger.debug(
            "TlsAlpnChallengeValidator._validate_certificate_extensions()"
        )
        try:
            from acme2certifier.acme_srv.helper import (
                cert_san_get,
                fqdn_in_san_check,
                cert_extensions_get,
            )
        except ImportError:
            self.logger.error(
                "Required helper functions not available for certificate validation"
            )
            return False

        san_list = cert_san_get(self.logger, cert, recode=False)
        fqdn_in_san = fqdn_in_san_check(self.logger, san_list, fqdn)

        if not fqdn_in_san:
            self.logger.debug(
                "TlsAlpnChallengeValidator._validate_certificate_extensions(): FQDN check against SAN failed"
            )
            return False

        extension_list = cert_extensions_get(self.logger, cert, recode=False)
        if extension_value in extension_list:
            self.logger.debug(
                "TlsAlpnChallengeValidator._validate_certificate_extensions(): TLS-ALPN validation successful"
            )
            return True
        else:
            self.logger.debug(
                "TlsAlpnChallengeValidator._validate_certificate_extensions(): TLS-ALPN validation not successful"
            )
            return False



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""unittests for account.py"""

# pylint: disable=C0302, C0415, R0904, R0913, R0914, R0915, W0212
import unittest
import sys
import configparser
from unittest.mock import patch, MagicMock

sys.path.insert(1, "..")


class FakeDBStore(object):
    """face DBStore class needed for mocking"""

    # pylint: disable=W0107, R0903
    pass


class TestACMEHandler(unittest.TestCase):
    """test class for ACMEHandler"""

    acme = None

    def setUp(self):
        """setup unittest"""
        models_mock = MagicMock()
        modules = {"acme2certifier.acme_srv.db_handler": models_mock}
        patch.dict("sys.modules", modules).start()
        import logging

        logging.basicConfig(level=logging.CRITICAL)
        self.logger = logging.getLogger("test_a2c")
        from acme2certifier.acme_srv.signature import Signature

        self.signature = Signature(False, "http://tester.local", self.logger)

    def test_001_signature__jwk_load(self):
        """test jwk load"""
        # Mock the dbstore instance on the existing signature object
        self.signature.dbstore = MagicMock()
        self.signature.dbstore.jwk_load.return_value = "foo"

        self.assertEqual("foo", self.signature._jwk_loader(1))

    def test_002_signature_check(self):
        """test Signature.check() without having content"""
        self.assertEqual(
            (False, "urn:ietf:params:acme:error:malformed", None),
            self.signature.check("foo", None),
        )

    @patch("acme2certifier.acme_srv.signature.Signature._jwk_loader")
    def test_003_signature_check(self, mock_jwk):
        """test Signature.check() while pubkey lookup failed"""
        mock_jwk.return_value = {}
        self.assertEqual(
            (False, "urn:ietf:params:acme:error:accountDoesNotExist", None),
            self.signature.check("foo", 1),
        )

    @patch("acme2certifier.acme_srv.signature.signature_check")
    @patch("acme2certifier.acme_srv.signature.Signature._jwk_loader")
    def test_004_signature_check(self, mock_jwk, mock_sig):
        """test successful Signature.check()"""
        mock_jwk.return_value = {"foo": "bar"}
        mock_sig.return_value = (True, None)
        self.assertEqual((True, None, None), self.signature.check("foo", 1))

    def test_005_signature_check(self):
        """test successful Signature.check() without account_name and use_emb_key False"""
        self.assertEqual(
            (False, "urn:ietf:params:acme:error:accountDoesNotExist", None),
            self.signature.check(None, 1, False),
        )

    def test_006_signature_check(self):
        """test successful Signature.check() without account_name and use_emb_key True but having a corrupted protected header"""
        protected = {"foo": "foo"}
        self.assertEqual(
            (False, "urn:ietf:params:acme:error:accountDoesNotExist", None),
            self.signature.check(None, 1, True, protected),
        )

    @patch("acme2certifier.acme_srv.signature.DBstore")
    @patch("acme2certifier.acme_srv.signature.signature_check")
    def test_007_signature_check(self, mock_sig, mock_dbstore_class):
        """test successful Signature.check() with account_name and use_emb_key True, sigcheck returns something"""
        # Setup dbstore mock to return a key
        mock_dbstore_instance = MagicMock()
        mock_dbstore_instance.jwk_load.return_value = {"key": "value"}
        mock_dbstore_class.return_value = mock_dbstore_instance

        # Setup signature_check mock
        mock_sig.return_value = ("result", "error")

        # Create a new signature instance with the mocked dbstore
        from acme2certifier.acme_srv.signature import Signature

        signature = Signature(False, "http://tester.local", self.logger)

        self.assertEqual(("result", "error", None), signature.check("foo", 1, True))

    @patch("acme2certifier.acme_srv.signature.signature_check")
    def test_008_signature_check(self, mock_sig):
        """test successful Signature.check() without account_name and use_emb_key True, sigcheck returns something"""
        mock_sig.return_value = ("result", "error")
        protected = {"url": "url", "jwk": "jwk"}
        self.assertEqual(
            ("result", "error", None), self.signature.check(None, 1, True, protected)
        )

    @patch("acme2certifier.acme_srv.signature.DBstore")
    def test_009_signature__jwk_load(self, mock_dbstore_class):
        """test jwk load  - dbstore.jwk_load() raises an exception"""
        # Setup mock to raise exception
        mock_dbstore_instance = MagicMock()
        mock_dbstore_instance.jwk_load.side_effect = Exception("exc_sig_jw_load")
        mock_dbstore_class.return_value = mock_dbstore_instance

        # Create a new signature instance with the mocked dbstore
        from acme2certifier.acme_srv.signature import Signature

        signature = Signature(False, "http://tester.local", self.logger)

        with self.assertLogs("test_a2c", level="INFO") as lcm:
            signature._jwk_loader(1)
        self.assertIn(
            "CRITICAL:test_a2c:Database error: failed to load JWK for account id 1: exc_sig_jw_load",
            lcm.output,
        )

    @patch("acme2certifier.acme_srv.signature.signature_check")
    def test_010_signature_eab_check(self, mock_sigchk):
        """test eab_check  -  result and error"""
        content = "content"
        mac_key = "mac_key"
        mock_sigchk.return_value = ("result", "error")
        self.assertEqual(
            ("result", "error"), self.signature.eab_check(content, mac_key)
        )

    @patch("acme2certifier.acme_srv.signature.signature_check")
    def test_011_signature_eab_check(self, mock_sigchk):
        """test eab_check  -  result no error"""
        content = "content"
        mac_key = "mac_key"
        mock_sigchk.return_value = ("result", None)
        self.assertEqual(("result", None), self.signature.eab_check(content, mac_key))

    @patch("acme2certifier.acme_srv.signature.signature_check")
    def test_012_signature_eab_check(self, mock_sigchk):
        """test eab_check  -  result false and  error"""
        content = "content"
        mac_key = "mac_key"
        mock_sigchk.return_value = (False, "error")
        self.assertEqual((False, "error"), self.signature.eab_check(content, mac_key))

    @patch("acme2certifier.acme_srv.signature.signature_check")
    def test_013_signature_eab_check(self, mock_sigchk):
        """test eab_check  -  content None"""
        content = None
        mac_key = "mac_key"
        mock_sigchk.return_value = (False, "error")
        self.assertEqual(
            (False, "urn:ietf:params:acme:error:malformed"),
            self.signature.eab_check(content, mac_key),
        )

    @patch("acme2certifier.acme_srv.signature.signature_check")
    def test_014_signature_eab_check(self, mock_sigchk):
        """test eab_check  -  mac_key None"""
        content = "content"
        mac_key = None
        mock_sigchk.return_value = (False, "error")
        self.assertEqual(
            (False, "urn:ietf:params:acme:error:malformed"),
            self.signature.eab_check(content, mac_key),
        )

    @patch("acme2certifier.acme_srv.signature.signature_check")
    def test_015_signature_eab_check(self, mock_sigchk):
        """test eab_check  -  mac_key and content None"""
        content = None
        mac_key = None
        mock_sigchk.return_value = (False, "error")
        self.assertEqual(
            (False, "urn:ietf:params:acme:error:malformed"),
            self.signature.eab_check(content, mac_key),
        )

    @patch("acme2certifier.acme_srv.signature.load_config")
    def test_016__init(self, mock_load_cfg):
        """test _config_load account with url prefix without tailing slash configured"""
        parser = configparser.ConfigParser()
        parser["Directory"] = {"foo": "bar", "url_prefix": "url_prefix"}
        mock_load_cfg.return_value = parser
        self.signature.__init__(False, "http://tester.local", self.logger)
        self.assertEqual("url_prefix/acme/revokecert", self.signature.revocation_path)

    @patch("acme2certifier.acme_srv.signature.load_config")
    def test_017__init(self, mock_load_cfg):
        """test _config_load account with url prefix without tailing slash configured"""
        parser = configparser.ConfigParser()
        parser["Directory"] = {"foo": "bar"}
        mock_load_cfg.return_value = parser
        self.signature.__init__(False, "http://tester.local", self.logger)
        self.assertEqual("/acme/revokecert", self.signature.revocation_path)

    @patch("acme2certifier.acme_srv.signature.load_config")
    def test_016__init_empty(self, mock_load_cfg):
        """test _config_load account with url prefix without tailing slash configured"""
        parser = configparser.ConfigParser()
        parser["CAHandler"] = {"foo": "bar"}
        mock_load_cfg.return_value = parser
        self.signature.__init__(False, "http://tester.local", self.logger)
        self.assertEqual("/acme/revokecert", self.signature.revocation_path)

    def test_018_signature_check(self):
        """test Signature.cli_check() without having aname"""
        self.assertEqual(
            (False, "urn:ietf:params:acme:error:accountDoesNotExist", None),
            self.signature.cli_check(None, "content"),
        )

    def test_019_signature_check(self):
        """test Signature.check() without having content"""
        self.assertEqual(
            (False, "urn:ietf:params:acme:error:malformed", None),
            self.signature.cli_check("foo", None),
        )

    @patch("acme2certifier.acme_srv.signature.Signature._jwk_loader")
    def test_020_signature_check(self, mock_jwk):
        """test Signature.check() while pubkey lookup failed"""
        mock_jwk.return_value = {}
        self.assertEqual(
            (False, "urn:ietf:params:acme:error:accountDoesNotExist", None),
            self.signature.cli_check("foo", 1),
        )

    @patch("acme2certifier.acme_srv.signature.signature_check")
    @patch("acme2certifier.acme_srv.signature.Signature._jwk_loader")
    def test_021_signature_check(self, mock_jwk, mock_sig):
        """test successful Signature.check()"""
        mock_jwk.return_value = {"foo": "bar"}
        mock_sig.return_value = (True, None)
        self.assertEqual((True, None, None), self.signature.cli_check("foo", 1))

    @patch("acme2certifier.acme_srv.signature.signature_check")
    @patch("acme2certifier.acme_srv.signature.Signature._jwk_loader")
    def test_022_signature_check(self, mock_jwk, mock_sig):
        """test successful Signature.check() without account_name  sigcheck returns something"""
        mock_jwk.return_value = {"foo": "bar"}
        mock_sig.return_value = ("result", "error")
        self.assertEqual(("result", "error", None), self.signature.cli_check("foo", 1))

    def test_023_cli_check_no_content(self):
        """Signature.cli_check() returns malformed error if content is None"""
        self.assertEqual(
            (False, "urn:ietf:params:acme:error:malformed", None),
            self.signature.cli_check("foo", None),
        )

    def test_024_cli_check_no_aname(self):
        """Signature.cli_check() returns accountDoesNotExist error if aname is None"""
        self.assertEqual(
            (False, "urn:ietf:params:acme:error:accountDoesNotExist", None),
            self.signature.cli_check(None, "content"),
        )

    @patch("acme2certifier.acme_srv.signature.Signature._jwk_loader")
    def test_025_cli_check_pubkey_none(self, mock_jwk):
        """Signature.cli_check() returns accountDoesNotExist error if pubkey is None"""
        mock_jwk.return_value = None
        self.assertEqual(
            (False, "urn:ietf:params:acme:error:accountDoesNotExist", None),
            self.signature.cli_check("foo", "content"),
        )

    @patch("acme2certifier.acme_srv.signature.signature_check")
    @patch("acme2certifier.acme_srv.signature.Signature._jwk_loader")
    def test_026_cli_check_success(self, mock_jwk, mock_sig):
        """Signature.cli_check() returns result from signature_check"""
        mock_jwk.return_value = {"foo": "bar"}
        mock_sig.return_value = (True, None)
        self.assertEqual((True, None, None), self.signature.cli_check("foo", "content"))

    def test_027_check_no_content(self):
        """Signature.check() returns malformed error if content is None"""
        self.assertEqual(
            (False, "urn:ietf:params:acme:error:malformed", None),
            self.signature.check("foo", None),
        )

    @patch("acme2certifier.acme_srv.signature.Signature._jwk_loader")
    def test_028_check_pubkey_none(self, mock_jwk):
        """Signature.check() returns accountDoesNotExist error if pubkey is None"""
        mock_jwk.return_value = None
        self.assertEqual(
            (False, "urn:ietf:params:acme:error:accountDoesNotExist", None),
            self.signature.check("foo", "content"),
        )

    @patch("acme2certifier.acme_srv.signature.signature_check")
    @patch("acme2certifier.acme_srv.signature.Signature._jwk_loader")
    def test_029_check_success(self, mock_jwk, mock_sig):
        """Signature.check() returns result from signature_check"""
        mock_jwk.return_value = {"foo": "bar"}
        mock_sig.return_value = (True, None)
        self.assertEqual((True, None, None), self.signature.check("foo", "content"))

    def test_030_check_emb_key_no_protected(self):
        """Signature.check() returns accountDoesNotExist error if use_emb_key True but protected is None"""
        self.assertEqual(
            (False, "urn:ietf:params:acme:error:accountDoesNotExist", None),
            self.signature.check(None, "content", True, None),
        )

    def test_031_check_emb_key_no_jwk(self):
        """Signature.check() returns accountDoesNotExist error if use_emb_key True but protected lacks jwk"""
        protected = {"foo": "bar"}
        self.assertEqual(
            (False, "urn:ietf:params:acme:error:accountDoesNotExist", None),
            self.signature.check(None, "content", True, protected),
        )

    @patch("acme2certifier.acme_srv.signature.signature_check")
    def test_032_check_emb_key_success(self, mock_sig):
        """Signature.check() returns result from signature_check with embedded jwk"""
        protected = {"jwk": {"foo": "bar"}}
        mock_sig.return_value = (True, None)
        self.assertEqual(
            (True, None, None), self.signature.check(None, "content", True, protected)
        )

    def test_033_check_no_aname_no_emb_key(self):
        """Signature.check() returns accountDoesNotExist error if no aname and use_emb_key False"""
        self.assertEqual(
            (False, "urn:ietf:params:acme:error:accountDoesNotExist", None),
            self.signature.check(None, "content", False),
        )

    def test_034_eab_check_no_content(self):
        """Signature.eab_check() returns malformed error if content is None"""
        self.assertEqual(
            (False, "urn:ietf:params:acme:error:malformed"),
            self.signature.eab_check(None, "mac_key"),
        )

    def test_035_eab_check_no_mac_key(self):
        """Signature.eab_check() returns malformed error if mac_key is None"""
        self.assertEqual(
            (False, "urn:ietf:params:acme:error:malformed"),
            self.signature.eab_check("content", None),
        )

    @patch("acme2certifier.acme_srv.signature.signature_check")
    def test_036_eab_check_success(self, mock_sig):
        """Signature.eab_check() returns result from signature_check"""
        mock_sig.return_value = (True, None)
        self.assertEqual((True, None), self.signature.eab_check("content", "mac_key"))

    @patch("acme2certifier.acme_srv.signature.signature_check")
    def test_037_eab_check_error(self, mock_sig):
        """Signature.eab_check() returns error from signature_check"""
        mock_sig.return_value = (False, "error")
        self.assertEqual(
            (False, "error"), self.signature.eab_check("content", "mac_key")
        )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""acmechallenge class"""

from __future__ import print_function
from acme2certifier.acme_srv.db_handler import DBstore


class Acmechallenge(object):
    """Acmechallenge handler"""

    def __init__(self, debug=None, srv_name=None, logger=None):
        self.server_name = srv_name
        self.debug = debug
        self.logger = logger
        self.dbstore = DBstore(self.debug, self.logger)

    def __enter__(self):
        """Makes ACMEHandler a Context Manager"""
        return self

    def __exit__(self, *args):
        """cose the connection at the end of the context"""

    def lookup(self, path_info: str) -> str:
        """check nonce"""
        self.logger.debug("Acmechallenge.lookup()")

        key_authorization = None
        if path_info:
            token = path_info.replace("/.well-known/acme-challenge/", "")
            self.logger.info("Lookup token: %s", token)
            challenge_dic = self.dbstore.cahandler_lookup("name", token)
            if challenge_dic and "value1" in challenge_dic:
                key_authorization = challenge_dic["value1"]

        self.logger.debug("Acmechallenge.lookup() ended with: %s", key_authorization)
        return key_authorization



"""rpc.py"""

# pylint: disable=C0209, C0415, E0401, R0913, W1201
import logging
from impacket import uuid
from impacket.dcerpc.v5 import epm, rpcrt, transport
from acme2certifier.cahandlers.ms_wcce.target import Target


def get_dce_rpc_from_string_binding(
    string_binding: str,
    target: Target,
    timeout: int = 5,
    target_ip: str = None,
    remote_name: str = None,
    auth_level: int = rpcrt.RPC_C_AUTHN_LEVEL_PKT_PRIVACY,
    do_kerberos=False,
) -> rpcrt.DCERPC_v5:
    """get dce from rpc"""
    if target_ip is None:
        target_ip = target.target_ip
    if remote_name is None:
        remote_name = target.remote_name

    target.do_kerberos = do_kerberos

    rpctransport = transport.DCERPCTransportFactory(string_binding)

    rpctransport.setRemoteHost(target_ip)
    rpctransport.setRemoteName(remote_name)

    rpctransport.set_connect_timeout(timeout)
    rpctransport.set_kerberos(target.do_kerberos, kdcHost=target.dc_ip)

    rpctransport.set_credentials(
        target.username,
        target.password,
        target.domain,
        target.lmhash,
        target.nthash,
        TGS=None,
    )

    dce = rpctransport.get_dce_rpc()
    dce.set_auth_level(auth_level)

    if target.do_kerberos is True:
        dce.set_auth_type(rpcrt.RPC_C_AUTHN_GSS_NEGOTIATE)

    return dce


def get_dynamic_endpoint(interface: bytes, target: str, timeout: int = 5):
    """get endpoint"""
    string_binding = r"ncacn_ip_tcp:%s[135]" % target
    rpctransport = transport.DCERPCTransportFactory(string_binding)
    rpctransport.set_connect_timeout(timeout)
    dce = rpctransport.get_dce_rpc()
    logging.debug(
        "Trying to resolve dynamic endpoint %s" % repr(uuid.bin_to_string(interface))
    )
    try:
        dce.connect()
    except Exception as err_:
        logging.warning("Failed to connect to endpoint mapper: %s" % err_)
        return None
    try:
        endpoint = epm.hept_map(target, interface, protocol="ncacn_ip_tcp", dce=dce)
        logging.debug(
            "Resolved dynamic endpoint %s to %s"
            % (repr(uuid.bin_to_string(interface)), repr(endpoint))
        )
        return endpoint
    except Exception:
        logging.debug(
            "Failed to resolve dynamic endpoint %s"
            % repr(uuid.bin_to_string(interface))
        )
        return None


def get_dce_rpc(
    interface: bytes,
    named_pipe: str,
    target: Target,
    timeout=5,
    verbose=False,
    do_kerberos=False,
    auth_level_np: int = rpcrt.RPC_C_AUTHN_LEVEL_PKT_PRIVACY,
    auth_level_dyn: int = rpcrt.RPC_C_AUTHN_LEVEL_PKT_PRIVACY,
) -> rpcrt.DCERPC_v5:
    """get dce rpc"""

    def _try_binding(string_binding: str, auth_level: int) -> rpcrt.DCERPC_v5:
        dce = get_dce_rpc_from_string_binding(
            string_binding,
            target,
            timeout,
            auth_level=auth_level,
            do_kerberos=do_kerberos,
        )
        logging.debug("Trying to connect to endpoint: %s" % string_binding)
        try:
            dce.connect()
        except Exception as err_:
            if verbose:
                logging.warning(
                    "Failed to connect to endpoint %s: %s" % (string_binding, err_)
                )
            return None

        logging.debug("Connected to endpoint: %s" % string_binding)

        dce.bind(interface)

        return dce

    def _try_np() -> rpcrt.DCERPC_v5:
        # Try named pipe
        string_binding = "ncacn_np:%s[%s]" % (target.target_ip, named_pipe)
        return _try_binding(string_binding, auth_level=auth_level_np)

    def _try_dyn() -> rpcrt.DCERPC_v5:
        string_binding = get_dynamic_endpoint(interface, target.target_ip, timeout)
        if string_binding is None:
            # Possible errors:
            # - TCP Port 135 is firewalled off
            # - CertSvc is not running
            logging.error("Failed to get dynamic TCP endpoint for CertSvc")
            return None

        dce = _try_binding(string_binding, auth_level=auth_level_dyn)
        return dce

    for method in [_try_np, _try_dyn]:
        dce = method()
        if dce is not None:
            return dce

    return None



# -*- coding: utf-8 -*-
"""Encoding and base64 utilities for acme2certifier"""

import base64
import textwrap
import logging


def b64decode_pad(logger: logging.Logger, string: str) -> bytes:
    """b64 decoding and padding of missing "=" """
    logger.debug("Helper.b64decode_pad()")
    try:
        b64dec = base64.urlsafe_b64decode(string + "=" * (4 - len(string) % 4))
    except Exception:
        b64dec = b"ERR: b64 decoding error"
    return b64dec.decode("utf-8")


def b64_decode(logger: logging.Logger, string: str) -> str:
    """b64 decoding"""
    logger.debug("Helper.b64decode()")
    return convert_byte_to_string(base64.b64decode(string))


def b64_encode(logger: logging.Logger, string: str) -> str:
    """encode a bytestream in base64"""
    logger.debug("Helper.b64_encode()")
    return convert_byte_to_string(base64.b64encode(string))


def b64_url_encode(logger: logging.Logger, string: str) -> str:
    """encode a bytestream in base64 url and remove padding"""
    logger.debug("Helper.b64_url_encode()")
    string = convert_string_to_byte(string)
    encoded = base64.urlsafe_b64encode(string)
    return encoded.rstrip(b"=")


def b64_url_recode(logger: logging.Logger, string: str) -> str:
    """recode base64_url to base64"""
    logger.debug("Helper.b64_url_recode()")
    padding_factor = (4 - len(string) % 4) % 4
    string = convert_byte_to_string(string)
    string += "=" * padding_factor
    result = str(string).translate(dict(zip(map(ord, "-_"), "+/")))
    return result


def b64_url_decode(logger: logging.Logger, string: str) -> str:
    """decode base64url encoded string"""
    logger.debug("Helper.b64_url_decode()")
    # Remove whitespace
    string = string.strip()
    # Add padding if missing
    pad = "=" * (-len(string) % 4)
    string_padded = string + pad
    return convert_byte_to_string(base64.urlsafe_b64decode(string_padded))


def build_pem_file(logger: logging.Logger, existing, certificate, wrap, csr=False):
    """construct pem_file"""
    logger.debug("Helper.build_pem_file()")
    if csr:
        pem_file = f"-----BEGIN CERTIFICATE REQUEST-----\n{textwrap.fill(convert_byte_to_string(certificate), 64)}\n-----END CERTIFICATE REQUEST-----\n"
    else:
        if existing:
            if wrap:
                pem_file = f"{existing}-----BEGIN CERTIFICATE-----\n{textwrap.fill(convert_byte_to_string(certificate), 64)}\n-----END CERTIFICATE-----\n"
            else:
                pem_file = f"{convert_byte_to_string(existing)}-----BEGIN CERTIFICATE-----\n{convert_byte_to_string(certificate)}\n-----END CERTIFICATE-----\n"
        else:
            if wrap:
                pem_file = f"-----BEGIN CERTIFICATE-----\n{textwrap.fill(convert_byte_to_string(certificate), 64)}\n-----END CERTIFICATE-----\n"
            else:
                pem_file = f"-----BEGIN CERTIFICATE-----\n{convert_byte_to_string(certificate)}\n-----END CERTIFICATE-----\n"
    return pem_file


def convert_byte_to_string(value: bytes) -> str:
    """convert a variable to string if needed"""
    if hasattr(value, "decode"):
        try:
            return value.decode()
        except Exception:
            return value
    else:
        return value


def convert_string_to_byte(value: str) -> bytes:
    """convert a variable to byte if needed"""
    if hasattr(value, "encode"):
        result = value.encode()
    else:
        result = value
    return result



#!/usr/bin/python
"""Invalidate expired orders/authorizations and refresh certificate dates."""

from __future__ import annotations

import time

from acme2certifier.acme_srv.helper import logger_setup, uts_to_date_utc
from acme2certifier.acme_srv.housekeeping import Housekeeping


def main() -> None:
    """Run housekeeping invalidation tasks."""
    debug = True
    logger = logger_setup(debug)
    suffix = uts_to_date_utc(int(time.time()), "%Y-%m-%d-%H%M%S")

    with Housekeeping(debug, logger) as housekeeping:
        housekeeping.orders_invalidate(
            report_format="csv", report_name=f"orders_invalidate_{suffix}"
        )
        housekeeping.authorizations_invalidate(
            report_format="csv", report_name=f"authorization_expire_{suffix}"
        )
        housekeeping.certificate_dates_update()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""unittests for acme2certifier"""

# pylint: disable= C0415, W0212
import unittest
import sys
import os
from unittest.mock import patch, MagicMock
import configparser

sys.path.insert(0, ".")
sys.path.insert(1, "..")


class TestEABHandler(unittest.TestCase):
    """test class for sql_handler"""

    def setUp(self):
        """setup unit test"""
        import sys
        import types

        sys.modules["psycopg2"] = types.ModuleType("psycopg2")
        sys.modules["psycopg2"].connect = MagicMock()
        mssql_mock = types.ModuleType("mssql_python")

        def dummy_connect(*args, **kwargs):
            return None

        mssql_mock.connect = dummy_connect
        sys.modules["mssql_python"] = mssql_mock
        import logging

        logging.basicConfig(level=logging.CRITICAL)
        self.logger = logging.getLogger("test_a2c")
        from acme2certifier.eabhandlers.sql_handler import EABhandler

        self.eabhandler = EABhandler(self.logger)
        self.dir_path = os.path.dirname(os.path.realpath(__file__))

    def test_001_default(self):
        """default test which always passes"""
        self.assertEqual("foo", "foo")

    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler._config_load")
    def test_002__enter__(self, mock_config_load):
        """test enter called"""
        mock_config_load.return_value = True
        self.eabhandler.__enter__()
        self.assertTrue(mock_config_load.called)

    @patch("acme2certifier.eabhandlers.sql_handler.load_config")
    def test_003_config_load(self, mock_config_load):
        """test _config_load - empty dictionary"""
        parser = configparser.ConfigParser()
        mock_config_load.return_value = parser
        self.eabhandler._config_load()
        self.assertFalse(self.eabhandler.db_system)
        self.assertFalse(self.eabhandler.db_host)
        self.assertFalse(self.eabhandler.db_name)
        self.assertFalse(self.eabhandler.db_user)
        self.assertFalse(self.eabhandler.db_password)

    @patch("acme2certifier.eabhandlers.sql_handler.load_config")
    def test_004_config_load(self, mock_load_config_load):
        """test _config_load - bogus values"""
        parser = configparser.ConfigParser()
        parser["foo"] = {"foo": "bar"}
        mock_load_config_load.return_value = parser
        self.eabhandler._config_load()
        self.assertFalse(self.eabhandler.db_system)
        self.assertFalse(self.eabhandler.db_host)
        self.assertFalse(self.eabhandler.db_name)
        self.assertFalse(self.eabhandler.db_user)
        self.assertFalse(self.eabhandler.db_password)

    @patch("acme2certifier.eabhandlers.sql_handler.load_config")
    def test_005_config_load(self, mock_config_load):
        """test _config_load - bogus values"""
        parser = configparser.ConfigParser()
        parser["EABhandler"] = {"foo": "bar"}
        mock_config_load.return_value = parser
        self.eabhandler._config_load()
        self.assertFalse(self.eabhandler.db_system)
        self.assertFalse(self.eabhandler.db_host)
        self.assertFalse(self.eabhandler.db_name)
        self.assertFalse(self.eabhandler.db_user)
        self.assertFalse(self.eabhandler.db_password)

    @patch("acme2certifier.eabhandlers.sql_handler.load_config")
    def test_006_config_load(self, mock_config_load):
        """test _config_load - valid values"""
        parser = configparser.ConfigParser()
        parser["EABhandler"] = {"db_system": "db_system"}
        mock_config_load.return_value = parser
        self.eabhandler._config_load()
        self.assertEqual("db_system", self.eabhandler.db_system)
        self.assertFalse(self.eabhandler.db_host)
        self.assertFalse(self.eabhandler.db_name)
        self.assertFalse(self.eabhandler.db_user)
        self.assertFalse(self.eabhandler.db_password)

    @patch("acme2certifier.eabhandlers.sql_handler.load_config")
    def test_007_config_load(self, mock_config_load):
        """test _config_load - valid values"""
        parser = configparser.ConfigParser()
        parser["EABhandler"] = {
            "db_system": "db_system",
            "db_host": "db_host",
            "db_name": "db_name",
            "db_user": "db_user",
            "db_password": "db_password",
        }
        mock_config_load.return_value = parser
        self.eabhandler._config_load()
        self.assertEqual("db_system", self.eabhandler.db_system)
        self.assertEqual("db_host", self.eabhandler.db_host)
        self.assertEqual("db_name", self.eabhandler.db_name)
        self.assertEqual("db_user", self.eabhandler.db_user)
        self.assertEqual("db_password", self.eabhandler.db_password)

    def test_008_mac_key_get(self):
        """test mac_key_get without db parameters specified"""
        self.assertFalse(self.eabhandler.mac_key_get(None))

    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler._wllist_check")
    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler._cn_add")
    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler._chk_san_lists_get")
    def test_009_allowed_domains_check(self, mock_san, mock_cn, mock_wlc):
        """test EABhanlder._allowed_domains_check()"""
        mock_san.return_value = (["foo"], [])
        mock_cn.return_value = ["foo", "bar"]
        mock_wlc.side_effect = [True, True]
        self.assertFalse(
            self.eabhandler._allowed_domains_check("csr", ["domain", "list"])
        )

    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler._wllist_check")
    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler._cn_add")
    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler._chk_san_lists_get")
    def test_010_allowed_domains_check(self, mock_san, mock_cn, mock_wlc):
        """test EABhanlder._allowed_domains_check()"""
        mock_san.return_value = (["foo"], [False])
        mock_cn.return_value = ["foo", "bar"]
        mock_wlc.side_effect = [True, True]
        self.assertEqual(
            "Either CN or SANs are not allowed by profile",
            self.eabhandler._allowed_domains_check("csr", ["domain", "list"]),
        )

    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler._wllist_check")
    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler._cn_add")
    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler._chk_san_lists_get")
    def test_011_allowed_domains_check(self, mock_san, mock_cn, mock_wlc):
        """test EABhanlder._allowed_domains_check()"""
        mock_san.return_value = (["foo"], [])
        mock_cn.return_value = ["foo", "bar"]
        mock_wlc.side_effect = [False, True]
        self.assertEqual(
            "Either CN or SANs are not allowed by profile",
            self.eabhandler._allowed_domains_check("csr", ["domain", "list"]),
        )

    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler.key_file_load")
    def test_012_eab_profile_get(self, mock_key_file_load):
        """test EABhandler._eab_profile_get()"""
        mock_key_file_load.return_value = {
            "eab_kid": {"cahandler": {"foo_parameter": "bar_parameter"}}
        }
        models_mock = MagicMock()
        models_mock.DBstore().certificate_lookup.return_value = {
            "foo": "bar",
            "order__account__eab_kid": "eab_kid",
        }
        modules = {"acme2certifier.acme_srv.db_handler": models_mock}
        patch.dict("sys.modules", modules).start()
        self.assertEqual(
            {"foo_parameter": "bar_parameter"}, self.eabhandler.eab_profile_get("csr")
        )

    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler.key_file_load")
    def test_013_eab_profile_get(self, mock_key_file_load):
        """test EABhandler._eab_profile_get()"""
        mock_key_file_load.return_value = {
            "eab_kid": {"cahandler_invalid": {"foo_parameter": "bar_parameter"}}
        }
        models_mock = MagicMock()
        models_mock.DBstore().certificate_lookup.return_value = {
            "foo": "bar",
            "order__account__eab_kid": "eab_kid",
        }
        modules = {"acme2certifier.acme_srv.db_handler": models_mock}
        patch.dict("sys.modules", modules).start()
        self.assertFalse(self.eabhandler.eab_profile_get("csr"))

    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler.key_file_load")
    def test_014_eab_profile_get(self, mock_key_file_load):
        """test EABhandler._eab_profile_get()"""
        mock_key_file_load.return_value = {
            "eab_kid": {"cahandler1": {"foo_parameter": "bar_parameter"}}
        }
        models_mock = MagicMock()
        models_mock.DBstore().certificate_lookup.return_value = {
            "foo": "bar",
            "1order__account__eab_kid": "eab_kid",
        }
        modules = {"acme2certifier.acme_srv.db_handler": models_mock}
        patch.dict("sys.modules", modules).start()
        self.assertFalse(self.eabhandler.eab_profile_get("csr"))

    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler.key_file_load")
    def test_015_eab_profile_get(self, mock_key_file_load):
        """test EABhandler._eab_profile_get()"""
        mock_key_file_load.return_value = {
            "eab_kid": {"cahandler": {"foo_parameter": "bar_parameter"}}
        }
        models_mock = MagicMock()
        models_mock.DBstore().certificate_lookup.return_value = {
            "foo": "bar",
            "order__account__eab_kid": "eab_kid1",
        }
        modules = {"acme2certifier.acme_srv.db_handler": models_mock}
        patch.dict("sys.modules", modules).start()
        self.assertFalse(self.eabhandler.eab_profile_get("csr"))

    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler.key_file_load")
    def test_016_eab_profile_get(self, mock_prof):
        """test EABhandler._eab_profile_get()"""
        mock_prof.return_value = {
            "eab_kid": {"cahandler": {"foo_parameter": "bar_parameter"}}
        }
        models_mock = MagicMock()
        models_mock.DBstore().certificate_lookup.side_effect = Exception("ex_db_lookup")
        modules = {"acme2certifier.acme_srv.db_handler": models_mock}
        patch.dict("sys.modules", modules).start()
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertFalse(self.eabhandler.eab_profile_get("csr"))
        self.assertIn(
            "ERROR:test_a2c:Database error while retrieving eab_kid: ex_db_lookup",
            lcm.output,
        )

    def test_017_chk_san_lists_get_empty(self):
        # Should return empty lists for empty input
        result = self.eabhandler._chk_san_lists_get(None)
        self.assertEqual(result, ([], []))

    @patch("acme2certifier.eabhandlers.sql_handler.csr_san_get")
    def test_018_chk_san_lists_get_value(self, mock_csr_san_get):
        # Should return empty lists for empty input
        mock_csr_san_get.return_value = ["dns:example.com", "dns:example.org"]
        result = self.eabhandler._chk_san_lists_get("csr")
        self.assertEqual(result, (["example.com", "example.org"], []))

    @patch("acme2certifier.eabhandlers.sql_handler.csr_san_get")
    def test_019_chk_san_lists_get_value(self, mock_csr_san_get):
        # Should return empty lists for empty input
        mock_csr_san_get.return_value = ["example.com", "example.org"]  #
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(
                [False, False], self.eabhandler._chk_san_lists_get("csr")[1]
            )
        self.assertIn(
            "INFO:test_a2c:SAN list parsing failed at entry: example.com", lcm.output
        )
        self.assertIn(
            "INFO:test_a2c:SAN list parsing failed at entry: example.org", lcm.output
        )

    @patch("acme2certifier.eabhandlers.sql_handler.csr_cn_get")
    def test_021_cn_add_cn_not_in_sans(self, mock_csr_cn_get):
        """CN present and not in SANs: should append CN"""
        mock_csr_cn_get.return_value = "example.com"
        result = self.eabhandler._cn_add("dummy_csr", ["test.com"])
        self.assertIn("example.com", result)
        self.assertIn("test.com", result)
        self.assertEqual(len(result), 2)

    @patch("acme2certifier.eabhandlers.sql_handler.csr_cn_get")
    def test_022_cn_add_cn_already_in_sans(self, mock_csr_cn_get):
        """CN present and already in SANs: should not duplicate CN"""
        mock_csr_cn_get.return_value = "example.com"
        result = self.eabhandler._cn_add("dummy_csr", ["example.com", "test.com"])
        self.assertIn("example.com", result)
        self.assertIn("test.com", result)
        self.assertEqual(len(result), 2)

    @patch("acme2certifier.eabhandlers.sql_handler.csr_cn_get")
    def test_023_cn_add_no_cn(self, mock_csr_cn_get):
        """No CN present: should not modify SANs"""
        mock_csr_cn_get.return_value = None
        result = self.eabhandler._cn_add("dummy_csr", ["test.com"])
        self.assertEqual(result, ["test.com"])

    def test_024_list_regex_check_match(self):
        """Entry matches regex: should return True"""
        result = self.eabhandler._list_regex_check("example.com", ["example\\.com"])
        self.assertTrue(result)

    def test_025_list_regex_check_no_match(self):
        """Entry does not match regex: should return False"""
        result = self.eabhandler._list_regex_check("example.com", ["test\\.com"])
        self.assertFalse(result)

    def test_026_list_regex_check_wildcard(self):
        """Entry matches wildcard regex: should return True"""
        result = self.eabhandler._list_regex_check(
            "sub.example.com", ["*.example\\.com"]
        )
        self.assertTrue(result)

    def test_027_wllist_check_match(self):
        """Entry matches list: should return True"""
        result = self.eabhandler._wllist_check("example.com", ["example\\.com"])
        self.assertTrue(result)

    def test_028_wllist_check_empty_list(self):
        """Empty list: should return True"""
        result = self.eabhandler._wllist_check("example.com", [])
        self.assertTrue(result)

    def test_029_wllist_check_toggle(self):
        """Toggle: should invert result"""
        result = self.eabhandler._wllist_check(
            "example.com", ["example\\.com"], toggle=True
        )
        self.assertFalse(result)

    def test_030_wllist_check_no_match(self):
        """Entry does not match list: should return False"""
        result = self.eabhandler._wllist_check("example.com", ["test\\.com"])
        self.assertFalse(result)

    def test_031_key_file_load_no_db_params(self):
        """No DB params: should return empty dict"""
        self.eabhandler.db_host = None
        self.eabhandler.db_name = None
        self.eabhandler.db_user = None
        self.eabhandler.db_password = None
        result = self.eabhandler.key_file_load()
        self.assertEqual(result, {})

    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler._load_mssql_profiles")
    def test_032_key_file_load_mssql(self, mock_load_mssql):
        """MSSQL: should call _load_mssql_profiles and return its result"""
        self.eabhandler.db_host = "host"
        self.eabhandler.db_name = "name"
        self.eabhandler.db_user = "user"
        self.eabhandler.db_password = "pass"
        self.eabhandler.db_system = "mssql"
        mock_load_mssql.return_value = {"key": "profile"}
        result = self.eabhandler.key_file_load()
        self.assertEqual(result, {"key": "profile"})
        mock_load_mssql.assert_called_once()

    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler._load_postgres_profiles")
    def test_033_key_file_load_postgres(self, mock_load_postgres):
        """Postgres: should call _load_postgres_profiles and return its result"""
        self.eabhandler.db_host = "host"
        self.eabhandler.db_name = "name"
        self.eabhandler.db_user = "user"
        self.eabhandler.db_password = "pass"
        self.eabhandler.db_system = "postgres"
        mock_load_postgres.return_value = {"key": "profile"}
        result = self.eabhandler.key_file_load()
        self.assertEqual(result, {"key": "profile"})
        mock_load_postgres.assert_called_once()

    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler._load_mssql_profiles")
    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler._load_postgres_profiles")
    def test_034_key_file_load_error(self, mock_postgres, mock_mssql):
        """Invalid db_system: should return empty dict"""
        self.eabhandler.db_host = "host"
        self.eabhandler.db_name = "name"
        self.eabhandler.db_user = "user"
        self.eabhandler.db_password = "pass"
        self.eabhandler.db_system = "invalid"
        mock_postgres.return_value = {}
        mock_mssql.return_value = {}
        result = self.eabhandler.key_file_load()
        self.assertEqual(result, {})

    @patch("acme2certifier.eabhandlers.sql_handler.connect")
    def test_035_load_mssql_profiles_success(self, mock_connect):
        """Successful fetch: should return dict with profiles"""
        self.eabhandler.db_host = "host"
        self.eabhandler.db_name = "name"
        self.eabhandler.db_user = "user"
        self.eabhandler.db_password = "pass"
        # Mock MSSQL connection and cursor
        mock_conn = MagicMock()
        mock_cursor = MagicMock()
        mock_conn.cursor.return_value = mock_cursor
        mock_cursor.fetchall.return_value = [
            MagicMock(key_id="id1", profile="profile1"),
            MagicMock(key_id="id2", profile="profile2"),
        ]
        mock_connect.return_value = mock_conn
        result = self.eabhandler._load_mssql_profiles("SELECT ...")
        self.assertEqual(result, {"id1": "profile1", "id2": "profile2"})
        mock_conn.close.assert_called_once()

    @patch("acme2certifier.eabhandlers.sql_handler.connect")
    def test_036_load_mssql_profiles_empty(self, mock_connect):
        """Empty result: should return empty dict"""
        mock_conn = MagicMock()
        mock_cursor = MagicMock()
        mock_conn.cursor.return_value = mock_cursor
        mock_cursor.fetchall.return_value = []
        mock_connect.return_value = mock_conn
        result = self.eabhandler._load_mssql_profiles("SELECT ...")
        self.assertEqual(result, {})

    @patch("acme2certifier.eabhandlers.sql_handler.connect")
    def test_037_load_mssql_profiles_exception(self, mock_connect):
        """Exception: should log error and return empty dict"""
        mock_connect.side_effect = Exception("connection error")
        with self.assertLogs("test_a2c", level="ERROR") as lcm:
            result = self.eabhandler._load_mssql_profiles("SELECT ...")
        self.assertEqual(result, {})
        self.assertTrue(any("error" in msg.lower() for msg in lcm.output))

    @patch("acme2certifier.eabhandlers.sql_handler.psycopg2.connect")
    def test_038_load_postgres_profiles_success(self, mock_connect):
        """Successful fetch: should return dict with profiles"""
        self.eabhandler.db_host = "host"
        self.eabhandler.db_name = "name"
        self.eabhandler.db_user = "user"
        self.eabhandler.db_password = "pass"
        mock_conn = MagicMock()
        mock_cursor = MagicMock()
        mock_conn.cursor.return_value = mock_cursor
        mock_cursor.fetchall.return_value = [("id1", "profile1"), ("id2", "profile2")]
        mock_connect.return_value = mock_conn
        result = self.eabhandler._load_postgres_profiles("SELECT ...")
        self.assertEqual(result, {"id1": "profile1", "id2": "profile2"})
        mock_conn.close.assert_called_once()

    @patch("acme2certifier.eabhandlers.sql_handler.psycopg2.connect")
    def test_039_load_postgres_profiles_empty(self, mock_connect):
        """Empty result: should return empty dict"""
        mock_conn = MagicMock()
        mock_conn.close = MagicMock()
        mock_conn.__bool__.return_value = True
        mock_cursor = MagicMock()
        mock_conn.cursor.return_value = mock_cursor
        mock_cursor.fetchall.return_value = []
        mock_connect.return_value = mock_conn
        result = self.eabhandler._load_postgres_profiles("SELECT ...")
        self.assertEqual(result, {})
        self.assertTrue(mock_conn.close.called)

    @patch("acme2certifier.eabhandlers.sql_handler.psycopg2.connect")
    def test_040_load_postgres_profiles_exception(self, mock_connect):
        """Exception: should log error and return empty dict"""
        mock_connect.side_effect = Exception("connection error")
        with self.assertLogs("test_a2c", level="ERROR") as lcm:
            result = self.eabhandler._load_postgres_profiles("SELECT ...")
        self.assertEqual(result, {})
        self.assertTrue(any("error" in msg.lower() for msg in lcm.output))

    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler.key_file_load")
    def test_041_mac_key_get_valid(self, mock_key_file_load):
        """Valid key: should return mac_key"""
        self.eabhandler.db_host = "host"
        self.eabhandler.db_name = "name"
        self.eabhandler.db_user = "user"
        self.eabhandler.db_password = "pass"
        mock_key_file_load.return_value = {"key1": "mac_value"}
        result = self.eabhandler.mac_key_get("key1")
        self.assertEqual(result, "mac_value")

    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler.key_file_load")
    def test_042_mac_key_get_missing_key(self, mock_key_file_load):
        """Missing key: should return None"""
        self.eabhandler.db_host = "host"
        self.eabhandler.db_name = "name"
        self.eabhandler.db_user = "user"
        self.eabhandler.db_password = "pass"
        mock_key_file_load.return_value = {"key1": "mac_value"}
        result = self.eabhandler.mac_key_get("key2")
        self.assertIsNone(result)

    def test_043_mac_key_get_missing_db_params(self):
        """Missing DB params: should return None and log error"""
        self.eabhandler.db_host = None
        self.eabhandler.db_name = None
        self.eabhandler.db_user = None
        self.eabhandler.db_password = None
        with self.assertLogs("test_a2c", level="ERROR") as lcm:
            result = self.eabhandler.mac_key_get("key1")
        self.assertIsNone(result)
        self.assertTrue(any("error" in msg.lower() for msg in lcm.output))

    @patch("acme2certifier.eabhandlers.sql_handler.EABhandler.key_file_load")
    def test_044_mac_key_get_exception(self, mock_key_file_load):
        """Exception: should return None and log error"""
        self.eabhandler.db_host = "host"
        self.eabhandler.db_name = "name"
        self.eabhandler.db_user = "user"
        self.eabhandler.db_password = "pass"
        mock_key_file_load.side_effect = Exception("lookup error")
        with self.assertLogs("test_a2c", level="ERROR") as lcm:
            result = self.eabhandler.mac_key_get("key1")
        self.assertIsNone(result)
        self.assertTrue(any("error" in msg.lower() for msg in lcm.output))

    def test_045_eab_kid_get_exception(self):
        """Exception branch: should log error and return None"""
        self.eabhandler.db_host = "host"
        self.eabhandler.db_name = "name"
        self.eabhandler.db_user = "user"
        self.eabhandler.db_password = "pass"
        dbstore_mock = MagicMock()
        dbstore_mock.side_effect = Exception("db error")
        sys.modules["acme2certifier.acme_srv.db_handler"] = MagicMock(
            DBstore=dbstore_mock
        )
        with self.assertLogs("test_a2c", level="ERROR") as lcm:
            result = self.eabhandler.eab_kid_get("csr")
        self.assertIsNone(result)
        self.assertTrue(
            any("Database error while retrieving eab_kid" in msg for msg in lcm.output)
        )


if __name__ == "__main__":

    if os.path.exists("acme_test.db"):
        os.remove("acme_test.db")
    unittest.main()
